"""command-line interface for messages package."""

import os

import click
from click import option

from messages import __version__ as VERSION
from .api import send
from ._config import create_config_profile
from ._config import CONFIG


##############################################################################
# Utility functions
##############################################################################


def get_body_from_file(kwds):
    """Reads message body if specified via filepath."""
    if kwds["file"] and os.path.isfile(kwds["file"]):
        kwds["body"] = open(kwds["file"], "r").read()
        kwds["file"] = None


def trim_args(kwds):
    """Gets rid of args with value of None, as well as select keys."""
    reject_key = ("type", "types", "configure")
    reject_val = (None, ())
    kwargs = {
        k: v for k, v in kwds.items() if k not in reject_key and v not in reject_val
    }
    for k, v in kwargs.items():
        if k in ("to", "cc", "bcc", "attachments"):
            kwargs[k] = list(kwargs[k])
    return kwargs


def send_message(msg_type, kwds):
    """Do some final preprocessing and send the message."""
    if kwds["file"]:
        get_body_from_file(kwds)
    kwargs = trim_args(kwds)
    send(msg_type, send_async=True, **kwargs)


##############################################################################
# Command Line Interface functions
##############################################################################


@click.group()
@click.version_option(version=VERSION, prog_name="Messages")
def main():
    """See available commands below to create appropriate message."""
    pass


@main.command("configure")
@click.argument("msg_type", required=True)
def main_configure(msg_type):
    """Configure profiles for the given message type."""
    create_config_profile(msg_type)


@main.command("email")
@click.argument("profile", type=click.STRING, required=False)
@click.argument("body", type=click.STRING, default="", required=False)
@option("-f", "--from", "from_", help="Originating (from) email address")
@option("-t", "--to", multiple=True, help="Recipient email addresses.")
@option(
    "-c", "--carboncopy", "cc", multiple=True, help="Carbon Copy (CC) email addresses."
)
@option(
    "-b",
    "--blindcopy",
    "bcc",
    multiple=True,
    help="Blind Carbon Copy (BCC) email addresses.",
)
@option("-s", "--subject", help="Subject line.")
@option("-m", "--file", help="Read message body from filepath.")
@option(
    "-a",
    "--attach",
    "attachments",
    multiple=True,
    help="Attachments -- filepath to attach.",
)
@option("-S", "--save", is_flag=True, help="Save default values in current profile.")
@option(
    "-V",
    "--verbose",
    is_flag=True,
    help="Display verbose output and debug information.",
)
@click.pass_context
def main_email(ctx, **kwds):
    """Send email message.

    * [PROFILE]: Pre-configured user profile.

    * [BODY]:    Message body text.

    * Example: messages email myEmailProfile 'Hello from email.' -t 'first@test.com' -c 'second@test.com' -s 'TestSubject'
    --attach ./sample.txt --verbose
    """
    send_message("email", kwds)


@main.command("twilio")
@click.argument("profile", type=click.STRING, required=False)
@click.argument("body", type=click.STRING, default="", required=False)
@option("-f", "--from", "from_", help="Originating (from) phone number.")
@option("-t", "--to", multiple=True, help="Recipient phone number.")
@option("-m", "--file", help="Read message body from filepath.")
@option(
    "-a",
    "--attach",
    "attachments",
    multiple=True,
    help="Attachments -- url for image to attach.",
)
@option("-S", "--save", is_flag=True, help="Save default values in current profile.")
@option(
    "-V",
    "--verbose",
    is_flag=True,
    help="Display verbose output and debug information.",
)
@click.pass_context
def main_twilio(ctx, **kwds):
    """Send twilio text message.

    * [PROFILE]: Pre-configured user profile.

    * [BODY]:    Message body text.

    * Example: messages twilio myTwilioProfile 'hello from twilio' -t '+12223334444' --verbose
    """
    send_message("twilio", kwds)


@main.command("slackwebhook")
@click.argument("profile", type=click.STRING, required=False)
@click.argument("body", type=click.STRING, default="", required=False)
@option("-m", "--file", help="Read message body from filepath.")
@option(
    "-a",
    "--attach",
    "attachments",
    multiple=True,
    help="Attachments -- url for image to attach.",
)
@option("-S", "--save", is_flag=True, help="Save default values in current profile.")
@option(
    "-V",
    "--verbose",
    is_flag=True,
    help="Display verbose output and debug information.",
)
@click.pass_context
def main_slackwebhook(ctx, **kwds):
    """Send SlackWebhook message.

    * [PROFILE]: Pre-configured user profile.

    * [BODY]:    Message body text.

    * Example: messages slackwebhook mySlackWebhookProfile "Hello from slackwebhook." -a "https://somewebresource.jpg" --verbose
    """
    send_message("slackwebhook", kwds)


@main.command("slackpost")
@click.argument("profile", type=click.STRING, required=False)
@click.argument("body", type=click.STRING, default="", required=False)
@option("-c", "--channel", help="Slack channel to post message to.")
@option("-m", "--file", help="Read message body from filepath.")
@option(
    "-a",
    "--attach",
    "attachments",
    multiple=True,
    help="Attachments -- url for image to attach.",
)
@option("-S", "--save", is_flag=True, help="Save default values in current profile.")
@option(
    "-V",
    "--verbose",
    is_flag=True,
    help="Display verbose output and debug information.",
)
@click.pass_context
def main_slackpost(ctx, **kwds):
    """Send SlackPost message.

    * [PROFILE]: Pre-configured user profile.

    * [BODY]:    Message body text.

    * Example: mesages slackpost mySlackProfile "Hello from Slack." -c "#general" --verbose
    """
    send_message("slackpost", kwds)


@main.command("telegram")
@click.argument("profile", type=click.STRING, required=False)
@click.argument("body", type=click.STRING, default="", required=False)
@option("-c", "--chat-id", help="Chat ID to send message to.")
@option("-m", "--file", help="Read message body from filepath.")
@option(
    "-a",
    "--attach",
    "attachments",
    multiple=True,
    help="Attachments -- url for image to attach.",
)
@option("-S", "--save", is_flag=True, help="Save default values in current profile.")
@option(
    "-V",
    "--verbose",
    is_flag=True,
    help="Display verbose output and debug information.",
)
@click.pass_context
def main_telegram(ctx, **kwds):
    """Send TelegramBot message.

    * [PROFILE]: Pre-configured user profile.

    * [BODY]:    Message body text.

    * Example: messages telegram myTelegramProfile "Hello from Telegram." -a "https://somefile.jpg" --verbose
    """
    send_message("telegrambot", kwds)


@main.command("whatsapp")
@click.argument("profile", type=click.STRING, required=False)
@click.argument("body", type=click.STRING, default="", required=False)
@option("-f", "--from", "from_", help="Originating (from) phone number.")
@option("-t", "--to", multiple=True, help="Recipient phone number.")
@option("-m", "--file", help="Read message body from filepath.")
@option(
    "-a",
    "--attach",
    "attachments",
    multiple=True,
    help="Attachments -- url for image to attach.",
)
@option("-S", "--save", is_flag=True, help="Save default values in current profile.")
@option(
    "-V",
    "--verbose",
    is_flag=True,
    help="Display verbose output and debug information.",
)
@click.pass_context
def main_whatsapp(ctx, **kwds):
    """Send WhatsApp text message via the Twilio API.

    * [PROFILE]: Pre-configured user profile.

    * [BODY]:    Message body text.

    * Example: messages whatsapp myWhatsAppProfile 'hello from whatsapp' -t '+12223334444' --verbose
    """
    send_message("whatsapp", kwds)



"""messages.eventloop tests."""

import pytest

import asyncio
from unittest.mock import Mock

from messages._eventloop import MessageLoop
from messages._exceptions import UnsupportedMessageTypeError


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_messageloop():
    """Return a valid MessageLoop object."""
    return MessageLoop()


class MsgGood:
    """Good because there is a send() method."""
    def __init__(self): pass
    def send(self): pass


class MsgBad:
    """Bad because there is no send() method."""
    def __init__(self): pass


##############################################################################
# TESTS: MessageLoop.__init__
##############################################################################

def test_init(get_messageloop):
    """
    GIVEN a need to create an MessageLoop object
    WHEN the user instantiates a new object
    THEN assert MessageLoop object is created with given args
    """
    ml = get_messageloop
    assert ml is not None


##############################################################################
# TESTS: MessageLoop.add_message
##############################################################################

def test_add_message_msgGood(get_messageloop, mocker):
    """
    GIVEN a valid MessageLoop object
    WHEN a valid message is added with the add_message method
    THEN assert it is added and send_loop() is called
    """
    send_loop_mock = mocker.patch.object(MessageLoop, 'send_loop')
    ml = get_messageloop
    ml.add_message(MsgGood())
    assert send_loop_mock.call_count == 1


def test_add_message_msgBad(get_messageloop):
    """
    GIVEN a valid MessageLoop object
    WHEN an invalid message is added with the add_message method
    THEN assert UnsupportedMessageTypeError is raised
    """
    ml = get_messageloop
    with pytest.raises(UnsupportedMessageTypeError):
        ml.add_message(MsgBad())


##############################################################################
# TESTS: MessageLoop.send_loop
##############################################################################

def test_send_loop_MessageGood(get_messageloop):
    """
    GIVEN a valid MessageLoop object
    WHEN a send_loop() is initiated with a valid message
    THEN assert the loop.run_in_executor is called to send the message
    """
    ml = get_messageloop
    ml.loop = Mock()
    ml.add_message(MsgGood())
    assert ml.loop.run_in_executor.call_count == 1



"""messages.slack tests."""

import pytest
import requests

import messages.slack
from messages.slack import SlackWebhook
from messages.slack import SlackPost
from messages._eventloop import MESSAGELOOP


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_slackWH(cfg_mock, mocker):
    """Return a valid SlackWebhook object."""
    mocker.patch.object(messages.slack, 'check_config_file')
    return SlackWebhook(auth='https://testurl.com', body='message',
            attachments=['https://url1.com', 'https://url2.com'],
            profile='myProfile')


@pytest.fixture()
def get_slackP(cfg_mock, mocker):
    """Return a valid SlackPost object."""
    mocker.patch.object(messages.slack, 'check_config_file')
    return SlackPost(auth='1234:ABCD', channel='general',
            body='message', attachments=['https://url1.com', 'https://url2.com'],
            profile='myProfile')


##############################################################################
# TESTS: Slack*.__init__
##############################################################################

def test_slackWH_init(get_slackWH):
    """
    GIVEN a need for a SlackWebhook object
    WHEN instantiated
    THEN assert it is properly created
    """
    s = get_slackWH
    assert s.body == 'message'
    assert s.auth == '***obfuscated***'
    assert '_auth' in s.__dict__
    assert s._auth == 'https://testurl.com'
    assert isinstance(s.message, dict)


def test_slackP_init(get_slackP):
    """
    GIVEN a need for a SlackPost object
    WHEN instantiated
    THEN assert it is properly created
    """
    s = get_slackP
    assert s.body == 'message'
    assert s.auth == '***obfuscated***'
    assert '_auth' in s.__dict__
    assert s._auth == '1234:ABCD'
    assert isinstance(s.message, dict)


##############################################################################
# TESTS: Slack*.__str__
##############################################################################

def test_slackWH_str(get_slackWH, capsys):
    """
    GIVEN a SlackWebhook object
    WHEN print() is called on the object
    THEN assert the proper output prints
    """
    s = get_slackWH
    print(s)
    out, err = capsys.readouterr()
    assert "['https://url1.com', 'https://url2.com']" in out
    assert "'message'" in out
    assert err == ''


def test_slackP_str(get_slackP, capsys):
    """
    GIVEN a SlackPost object
    WHEN print() is called on the object
    THEN assert the proper output prints
    """
    s = get_slackP
    print(s)
    out, err = capsys.readouterr()
    assert "['https://url1.com', 'https://url2.com']" in out
    assert "'message'" in out
    assert 'gener' in out
    assert err == ''


##############################################################################
# TESTS: Slack*._construct_message
##############################################################################

def test_slackWH_construct_message(get_slackWH, mocker):
    """
    GIVEN a valid SlackWebhook object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    add_mock = mocker.patch.object(SlackWebhook, '_add_attachments')
    s = get_slackWH
    s._construct_message()
    assert s.message['text'] == 'message'
    assert add_mock.call_count == 1


def test_slackP_construct_message(get_slackP, mocker):
    """
    GIVEN a valid SlackPost object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    add_mock = mocker.patch.object(SlackPost, '_add_attachments')
    s = get_slackP
    s._construct_message()
    assert s.message['text'] == 'message'
    assert add_mock.call_count == 1


def test_slackWH_construct_message_withFromSubj(get_slackWH, mocker):
    """
    GIVEN a valid SlackWebhook object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    add_mock = mocker.patch.object(SlackWebhook, '_add_attachments')
    s = get_slackWH
    s.from_ = 'me'
    s.subject = 'Tst Msg'
    s._construct_message()
    expected = 'From: me\nSubject: Tst Msg\nmessage'
    assert s.message['text'] == expected
    assert add_mock.call_count == 1


def test_slackP_construct_message_withFromSubj(get_slackP, mocker):
    """
    GIVEN a valid SlackPost object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    add_mock = mocker.patch.object(SlackPost, '_add_attachments')
    s = get_slackP
    s.from_ = 'me'
    s.subject = 'Tst Msg'
    s._construct_message()
    expected = 'From: me\nSubject: Tst Msg\nmessage'
    assert s.message['text'] == expected
    assert add_mock.call_count == 1


##############################################################################
# TESTS: Slack*._add_attachments
##############################################################################

def test_slackWH_add_attachments_list(get_slackWH):
    """
    GIVEN a valid SlackWebhook object with self.attach_urls = list
    WHEN add_attachments() is called
    THEN assert the urls are properly attached to the message
    """
    s = get_slackWH
    s._add_attachments()
    expected = [{'image_url': 'https://url1.com', 'author_name': ''},
                {'image_url': 'https://url2.com', 'author_name': ''}]
    assert s.message['attachments'] == expected


def test_slackP_add_attachments_list(get_slackP):
    """
    GIVEN a valid SlackPost object with self.attach_urls = list
    WHEN add_attachments() is called
    THEN assert the urls are properly attached to the message
    """
    s = get_slackP
    s._add_attachments()
    expected = [{'image_url': 'https://url1.com', 'author_name': ''},
                {'image_url': 'https://url2.com', 'author_name': ''}]
    assert s.message['attachments'] == expected


def test_slackWH_add_attachments_str(get_slackWH):
    """
    GIVEN a valid SlackWebhook object with self.attach_urls = str
    WHEN add_attachments() is called
    THEN assert the urls are properly attached to the message
    """
    s = get_slackWH
    s.attachments = 'https://url1.com'
    s._add_attachments()
    expected = [{'image_url': 'https://url1.com', 'author_name': ''}]
    assert s.message['attachments'] == expected


def test_slackP_add_attachments_str(get_slackP):
    """
    GIVEN a valid SlackPost object with self.attach_urls = str
    WHEN add_attachments() is called
    THEN assert the urls are properly attached to the message
    """
    s = get_slackP
    s.attachments = 'https://url1.com'
    s._add_attachments()
    expected = [{'image_url': 'https://url1.com', 'author_name': ''}]
    assert s.message['attachments'] == expected


def test_slackWH_add_attachments_with_params(get_slackWH):
    """
    GIVEN a valid SlackWebhook object with extra attachment params
    WHEN add_attachments() is called
    THEN assert the extra params are properly added
    """
    s = get_slackWH
    s.attachments = 'https://url1.com'
    s.params = {'author_name': 'me', 'text': 'image of me'}
    s._add_attachments()
    expected = [{'image_url': 'https://url1.com', 'author_name': 'me',
                'text': 'image of me'}]
    assert s.message['attachments'] == expected


def test_slackP_add_attachments_with_params(get_slackP):
    """
    GIVEN a valid SlackPost object with extra attachment params
    WHEN add_attachments() is called
    THEN assert the extra params are properly added
    """
    s = get_slackP
    s.attachments = 'https://url1.com'
    s.params = {'author_name': 'me', 'text': 'image of me'}
    s._add_attachments()
    expected = [{'image_url': 'https://url1.com', 'author_name': 'me',
                'text': 'image of me'}]
    assert s.message['attachments'] == expected


##############################################################################
# TESTS: Slack*.send
##############################################################################

def test_slackWH_send_status201(get_slackWH, capsys, mocker):
    """
    GIVEN a valid SlackWebhook object
    WHEN send() is called
    THEN assert the proper send sequence occurs
    """
    con_mock = mocker.patch.object(SlackWebhook, '_construct_message')
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 201
    s = get_slackWH
    s.send()
    out, err = capsys.readouterr()
    assert con_mock.call_count == 1
    assert req_mock.call_count == 1
    assert out == 'Message sent.\n'
    assert err == ''


def test_slackWH_send_status301(get_slackWH, capsys, mocker):
    """
    GIVEN a valid SlackWebhook object
    WHEN send() is called
    THEN assert the proper send sequence occurs
    """
    con_mock = mocker.patch.object(SlackWebhook, '_construct_message')
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 301
    s = get_slackWH
    s.send()
    out, err = capsys.readouterr()
    assert con_mock.call_count == 1
    assert req_mock.call_count == 1
    assert 'Error while sending.  HTTP status code = 301' in out
    assert err == ''


def test_slackP_send_status_201(get_slackP, capsys, mocker):
    """
    GIVEN a valid SlackPost object
    WHEN send() is called
    THEN assert the proper send sequence occurs
    """
    con_mock = mocker.patch.object(SlackPost, '_construct_message')
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 201
    s = get_slackP
    s.send()
    out, err = capsys.readouterr()
    assert con_mock.call_count == 1
    assert req_mock.call_count == 1
    assert out == 'Message sent.\n'
    assert err == ''


def test_slackWH_send_verbose_true(get_slackWH, capsys, mocker):
    """
    GIVEN a valid SlackWebhook object
    WHEN *.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to True)
    """
    con_mock = mocker.patch.object(SlackWebhook, '_construct_message')
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 201
    s = get_slackWH
    s.verbose = True
    s.send()
    out, err = capsys.readouterr()
    assert 'Debugging info' in out
    assert 'Message created.' in out
    assert ' * URL: https://testurl.com' in out
    assert ' * From: Not Specified' in out
    assert ' * Subject: None' in out
    assert ' * Body: \'message\'' in out
    assert ' * Attachments: [\'https://url1.com\', \'https://url2.com\']' in out
    assert ' * HTTP status code: 201' in out
    assert 'Message sent.' in out
    assert err == ''


def test_slackWH_send_verbose_false(get_slackWH, capsys, mocker):
    """
    GIVEN a valid SlackWebhook object
    WHEN *.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to False)
    """
    con_mock = mocker.patch.object(SlackWebhook, '_construct_message')
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 201
    s = get_slackWH
    s.verbose = False
    s.send()
    out, err = capsys.readouterr()
    assert 'Debugging info' not in out
    assert 'Message created.' not in out
    assert ' * URL: https://test_url.com' not in out
    assert ' * From: Not Specified' not in out
    assert ' * Subject: None' not in out
    assert ' * Body: \'message\'' not in out
    assert ' * Attachments: [\'https://url1.com\', \'https://url2.com\']' not in out
    assert 'Message sent.' in out
    assert err == ''


def test_slackP_send_verbose_true(get_slackP, capsys, mocker):
    """
    GIVEN a valid SlackPost object
    WHEN *.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to True)
    """
    con_mock = mocker.patch.object(SlackPost, '_construct_message')
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 201
    s = get_slackP
    s.verbose = True
    s.send()
    out, err = capsys.readouterr()
    assert 'Debugging info' in out
    assert 'Message created.' in out
    assert ' * Channel: general' in out
    assert ' * From: Not Specified' in out
    assert ' * Subject: None' in out
    assert ' * Body: \'message\'' in out
    assert ' * Attachments: [\'https://url1.com\', \'https://url2.com\']' in out
    assert 'Message sent.' in out
    assert err == ''


def test_slackP_send_verbose_false(get_slackP, capsys, mocker):
    """
    GIVEN a valid SlackPost object
    WHEN *.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to False)
    """
    con_mock = mocker.patch.object(SlackPost, '_construct_message')
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 201
    s = get_slackP
    s.verbose = False
    s.send()
    out, err = capsys.readouterr()
    assert 'Debugging info' not in out
    assert 'Message created.' not in out
    assert ' * Channel: general' not in out
    assert ' * From: Not Specified' not in out
    assert ' * Subject: None' not in out
    assert ' * Body: \'message\'' not in out
    assert ' * Attachments: [\'https://url1.com\', \'https://url2.com\']' not in out
    assert 'Message sent.' in out
    assert err == ''

##############################################################################
# TESTS: Slack*.send_async
##############################################################################

def test_slackWH_send_async(get_slackWH, mocker):
    """
    GIVEN a valid SlackWebhook object
    WHEN send_async() is called
    THEN assert the proper send sequence occurs
    """
    msgloop_mock = mocker.patch.object(MESSAGELOOP, 'add_message')
    s = get_slackWH
    s.send_async()
    assert msgloop_mock.call_count == 1


def test_slackP_send_async(get_slackP, mocker):
    """
    GIVEN a valid SlackPost object
    WHEN send_async() is called
    THEN assert the proper send sequence occurs
    """
    msgloop_mock = mocker.patch.object(MESSAGELOOP, 'add_message')
    s = get_slackP
    s.send_async()
    assert msgloop_mock.call_count == 1



"""Interfaces for message classes."""

import reprlib
from abc import ABCMeta
from abc import abstractmethod


class Message(metaclass=ABCMeta):
    """Interface for standard message classes."""

    @abstractmethod
    def send(self):
        """Send message synchronously."""

    @abstractmethod
    def send_async(self):
        """Send message asynchronously."""

    def __repr__(self):
        """repr(self) in debugging format with auth attr obfuscated."""
        class_name = type(self).__name__
        output = "{}(\n".format(class_name)
        for attr in self:
            if attr == "_auth":
                output += "auth=***obfuscated***,\n"
            elif attr == "body":
                output += "{}={!r},\n".format(attr, reprlib.repr(getattr(self, attr)))
            else:
                output += "{}={!r},\n".format(attr, getattr(self, attr))
        output += ")"
        return output

    def __iter__(self):
        return iter(self.__dict__)



"""messages._exceptions tests."""

import pytest

from messages._exceptions import InvalidMessageInputError
from messages._exceptions import UnsupportedMessageTypeError
from messages._exceptions import UnknownProfileError


##############################################################################
# TESTS: MessageInputError
##############################################################################

def test_InvalidMessageInputError(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(InvalidMessageInputError):
        raise InvalidMessageInputError('Email', 'from_', 'some_value', 'email address')
        out, err = capsys.readouterr()
        assert 'Invalid input for specified message class: Email' in out
        assert '* argument: "from_"' in out
        assert '* value given: some_value'
        assert '* input type must be: email address' in out
        assert err == ''


def test_UnsupportedMessageTypeError_default(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(UnsupportedMessageTypeError):
        raise UnsupportedMessageTypeError('BadType')
        out, err = capsys.readouterr()
        assert 'Invalid input for specified message class: BadType' in out
        assert err == ''


def test_UnsupportedMessageTypeError_listarg(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(UnsupportedMessageTypeError):
        raise UnsupportedMessageTypeError('BadType', {'m1', 'm2'})
        out, err = capsys.readouterr()
        assert 'Invalid input for specified message class: BadType' in out
        assert '* Supported message types:' in out
        assert "{'m1', 'm2'}" in out
        assert err == ''


def test_UnknownProfileError(capsys):
    """
    GIVEN no object instantiated, just raise the exception
    WHEN the exception is raised with given args
    THEN assert it raises exception and prints proper output
    """
    with pytest.raises(UnknownProfileError):
        raise UnknownProfileError('unknown_user')
        out, err = capsys.readouterr()
        assert 'Unknown Profile name: unknown_user' in out
        assert err == ''



"""reusable testing fixtures."""

import os
import pytest

import jsonconfig


##############################################################################
# Travis CI fixtures
##############################################################################

# skip this test if on travs-ci
skip_if_on_travisCI = pytest.mark.skipif("TRAVIS" in os.environ and
                    os.environ["TRAVIS"] == "true",
                    reason='skipping test if on travis-ci')


# skip this test if NOT on travis-ci
skip_if_not_on_travisCI = pytest.mark.skipif("TRAVIS" not in os.environ,
                    reason='skipping test if not on travis-ci')


##############################################################################
# jsonconfig fixtures
##############################################################################

class ConfigMock:
    """Mock class to return instead of jsonconfig.Config."""
    data = {'tester': 'this is a mock'}
    filename = '/default/path'
    pwd = {'myProf_email': None}
    kwargs = {'dump': {'indent': 4}}


@pytest.fixture()
def cfg_mock(monkeypatch):
    """Patches the context manager call `with jsonconfig.Config...`."""
    def enter(param1):
        return ConfigMock()

    def exit(param1, param2, param3, param4):
        pass

    monkeypatch.setattr(jsonconfig.Config, '__enter__', enter)
    monkeypatch.setattr(jsonconfig.Config, '__exit__', exit)



"""
Module that will handle asynchronous message sending, so each message will
be non-blocking.
"""

import asyncio

from ._exceptions import UnsupportedMessageTypeError


class MessageLoop:
    """Asynchronous message sending loop."""

    def __init__(self):
        self.loop = asyncio.get_event_loop()

    def add_message(self, message):
        """Add a message to the event loop."""
        try:
            self.send_loop(message)
        except AttributeError:
            raise UnsupportedMessageTypeError(message.__class__.__name__)

    def send_loop(self, msg, executor=None):
        """Send the message via the event loop."""
        self.loop.run_in_executor(executor, msg.send)


MESSAGELOOP = MessageLoop()



"""messages.email_ tests."""

import getpass
import os
import pytest
import smtplib

from email.mime.multipart import MIMEMultipart

import messages
from messages.email_ import Email
from messages.email_ import check_config_file
from messages._eventloop import MESSAGELOOP

from conftest import skip_if_on_travisCI
from conftest import skip_if_not_on_travisCI

##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_email(mocker):
    """Return a valid Email object."""
    config_mock = mocker.patch.object(messages.email_, 'check_config_file')
    e = Email(from_='me@here.com', to='you@there.com',
            server='smtp.gmail.com', port=465, auth='password',
            cc='someone@there.com', bcc='them@there.com',
            subject='subject', body='message', attachments=['file1', 'file2'],
            profile='myName', save=False)
    e.from_ = 'me@here.com'
    e.server = 'smtp.gmail.com'
    e.port = 465
    e.auth = 'password'
    return e


##############################################################################
# TESTS: Email.__init__
##############################################################################

def test_email_init_normal(get_email):
    """
    GIVEN a need to create an Email object
    WHEN the user instantiates a new object with required args
    THEN assert Email object is created with given args
    """
    e = get_email
    assert e is not None
    assert e.server == 'smtp.gmail.com'
    assert e.port == 465
    assert e.auth == '***obfuscated***'
    assert e.from_ == 'me@here.com'
    assert e.to == 'you@there.com'
    assert e.cc == 'someone@there.com'
    assert e.bcc == 'them@there.com'
    assert e.subject == 'subject'
    assert e.body == 'message'
    assert e.attachments == ['file1', 'file2']
    assert e.message is None


##############################################################################
# TESTS: Email.__str__
##############################################################################

def test_email_str(get_email, capsys):
    """
    GIVEN a valid Email object
    WHEN the user calls print() on the Email object
    THEN assert the correct format prints
    """
    e = get_email
    expected = ('\nServer: smtp.gmail.com:465'
                '\nFrom: me@here.com'
                '\nTo: you@there.com'
                '\nCc: someone@there.com'
                '\nBcc: them@there.com'
                '\nSubject: subject'
                '\nBody: \'message\''
                '\nAttachments: [\'file1\', \'file2\']\n')
    print(e)
    out, err = capsys.readouterr()
    assert out == expected
    assert err == ''


##############################################################################
# TESTS: Email.get_server
##############################################################################

def test_get_server_known(get_email):
    """
    GIVEN a valid Email object
    WHEN get_server() is called with an email address
    THEN assert a known smtp server is returned
    """
    e = get_email
    server = e.get_server('me@gmail.com')
    assert server == ('smtp.gmail.com', 465)


def test_get_server_guess(get_email):
    """
    GIVEN a valid Email object
    WHEN get_server() is called with an email address
    THEN assert an smtp server guess is returned
    """
    e = get_email
    server = e.get_server('me@test.com')
    assert server == ('smtp.test.com', 465)


def test_get_server_noAddress(get_email):
    """
    GIVEN a valid Email object
    WHEN get_server() is called without an email address
    THEN assert (None, None) is returned
    """
    e = get_email
    e.from_ = None
    server = e.get_server(e.from_)
    assert server == (None, None)


##############################################################################
# TESTS: Email.list_to_string
##############################################################################

def test_list_to_string(get_email):
    """
    GIVEN a valid Email object
    WHEN given a list of recipients
    THEN assert the list is returned as a single string
    """
    e = get_email
    output = e.list_to_string(['her@there.com', 'him@there.com'])
    assert output == 'her@there.com, him@there.com'


##############################################################################
# TESTS: Email.generate_email
##############################################################################

def test_generate_email(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.generate_email() is called
    THEN assert the email structure is created
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    attach_mock = mocker.patch.object(Email, '_add_attachments')
    e = get_email
    e._generate_email()
    assert isinstance(e.message, MIMEMultipart)
    assert header_mock.call_count == 1
    assert body_mock.call_count == 1
    assert attach_mock.call_count == 1


##############################################################################
# TESTS: Email.add_header
##############################################################################

def test_add_header(get_email, mocker):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
    WHEN Email.add_header() is called
    THEN assert correct parameters are set
    """
    body_mock = mocker.patch.object(Email, '_add_body')
    attach_mock = mocker.patch.object(Email, '_add_attachments')
    e = get_email
    e._generate_email()
    assert e.message['From'] == 'me@here.com'
    assert e.message['Subject'] == 'subject'


##############################################################################
# TESTS: Email.add_body
##############################################################################

def test_add_body(get_email, mocker):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
    WHEN Email.add_body() is called
    THEN assert body_text is attached
    """
    attach_mock = mocker.patch.object(Email, '_add_attachments')
    header_mock = mocker.patch.object(Email, '_add_header')
    mime_attach_mock = mocker.patch.object(MIMEMultipart, 'attach')
    e = get_email
    e._generate_email()
    assert mime_attach_mock.call_count == 1


##############################################################################
# TESTS: Email.add_attachments
##############################################################################

@skip_if_on_travisCI
def test_add_attachments_list_local(get_email, mocker):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
        and Email.attachments is a list
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    mime_attach_mock = mocker.patch.object(MIMEMultipart, 'attach')
    e = get_email
    e.attachments = ['tests/data/file1.txt', 'tests/data/file2.png',
                     'tests/data/file3.pdf', 'tests/data/file4.xlsx']
    e._generate_email()
    assert mime_attach_mock.call_count == 4


@skip_if_not_on_travisCI
def test_add_attachments_list_travis(get_email, mocker):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
         and Email.attachments is a list
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    mime_attach_mock = mocker.patch.object(MIMEMultipart, 'attach')
    e = get_email
    PATH = '/home/travis/build/trp07/messages/tests/data/'
    e.attachments = [PATH + 'file1.txt', PATH + 'file2.png',
                     PATH + 'file3.pdf', PATH + 'file4.xlsx']
    e._generate_email()
    assert mime_attach_mock.call_count == 4


@skip_if_on_travisCI
def test_add_attachments_str_local(get_email, mocker):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
        and Email.attachments is a str
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    mime_attach_mock = mocker.patch.object(MIMEMultipart, 'attach')
    e = get_email
    e.attachments = 'tests/data/file1.txt'
    e._generate_email()
    assert mime_attach_mock.call_count == 1


@skip_if_not_on_travisCI
def test_add_attachments_str_travis(get_email, mocker):
    """
    GIVEN a valid Email object, where Email.generate_email() has been called
         and Email.attachments is a str
    WHEN Email.add_attachments() is called
    THEN assert correct attachments are attached
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    mime_attach_mock = mocker.patch.object(MIMEMultipart, 'attach')
    e = get_email
    PATH = '/home/travis/build/trp07/messages/tests/data/'
    e.attachments = PATH + 'file1.txt'
    e._generate_email()
    assert mime_attach_mock.call_count == 1


##############################################################################
# TESTS: Email.get_session
##############################################################################

def test_get_session_ssl(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.get_session() is called
    THEN assert the correct functions are called
    """
    getssl_mock = mocker.patch.object(Email, '_get_ssl')
    e = get_email
    e._get_session()
    assert getssl_mock.call_count == 1


def test_get_session_tls(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.get_session() is called
    THEN assert the correct functions are called
    """
    gettls_mock = mocker.patch.object(Email, '_get_tls')
    e = get_email
    e.port = 587
    e._get_session()
    assert gettls_mock.call_count == 1


##############################################################################
# TESTS: Email.get_ssl
##############################################################################

def test_get_ssl(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.get_ssl() is called
    THEN assert an SMTP_SSL instance is invoked
    """
    smtpssl_mock = mocker.patch.object(smtplib, 'SMTP_SSL')
    e = get_email
    e._get_ssl()
    assert smtpssl_mock.call_count == 1


def test_get_ssl_port_string(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN port type is set to string
    THEN assert the correct functions are called
    """
    getssl_mock = mocker.patch.object(Email, '_get_ssl')
    e = get_email
    e.port = '465'
    e._get_session()
    assert getssl_mock.call_count == 1


##############################################################################
# TESTS: Email.get_tls
##############################################################################

def test_get_tls(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.get_tls() is called
    THEN assert an SMTP instance is invoked
    """
    smtp_mock = mocker.patch.object(smtplib, 'SMTP')
    e = get_email
    e._get_tls()
    assert smtp_mock.call_count == 1


def test_get_tls_port_string(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN port type is set to string
    THEN assert the correct functions are called
    """
    gettls_mock = mocker.patch.object(Email, '_get_tls')
    e = get_email
    e.port = '587'
    e._get_session()
    assert gettls_mock.call_count == 1


##############################################################################
# TESTS: Email.send
##############################################################################

def test_send(get_email, capsys, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.send() is called
    THEN assert the correct functions are called and correct attributes
        updated
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    attach_mock = mocker.patch.object(Email, '_add_attachments')
    session_mock = mocker.patch.object(Email, '_get_session')
    e = get_email
    e.send()
    out, err = capsys.readouterr()
    assert out == 'Message sent.\n'
    assert err == ''


def test_send_verbose_true(get_email, capsys, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to True)
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    attach_mock = mocker.patch.object(Email, '_add_attachments')
    session_mock = mocker.patch.object(Email, '_get_session')
    e = get_email
    e.verbose = True
    e.send()
    out, err = capsys.readouterr()
    assert 'Debugging info' in out
    assert 'Message created.' in out
    assert 'Login successful.' in out
    assert 'Logged out.' in out
    assert ' * Server: smtp.gmail.com:465' in out
    assert ' * From: me@here.com' in out
    assert ' * To: you@there.com' in out
    assert ' * Cc: someone@there.com' in out
    assert ' * Bcc: them@there.com' in out
    assert ' * Subject: subject' in out
    assert ' * Body: \'message\'' in out
    assert ' * Attachments: [\'file1\', \'file2\']' in out
    assert 'Message sent.' in out
    assert err == ''


def test_send_verbose_false(get_email, capsys, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to False)
    """
    header_mock = mocker.patch.object(Email, '_add_header')
    body_mock = mocker.patch.object(Email, '_add_body')
    attach_mock = mocker.patch.object(Email, '_add_attachments')
    session_mock = mocker.patch.object(Email, '_get_session')
    e = get_email
    e.verbose = False
    e.send()
    out, err = capsys.readouterr()

    assert out == 'Message sent.\n'
    assert 'Debugging info' not in out
    assert 'Message created.' not in out
    assert 'Login successful.' not in out
    assert 'Logged out.' not in out
    assert ' * Server: smtp.gmail.com:465' not in out
    assert ' * From: me@here.com' not in out
    assert err == ''


##############################################################################
# TESTS: Email.send_async
##############################################################################

def test_send_async(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.send_async() is called
    THEN assert it is added to the event loop for async sending
    """
    msg_loop_mock = mocker.patch.object(MESSAGELOOP, 'add_message')
    e = get_email
    e.send_async()
    assert msg_loop_mock.call_count == 1


def test_send_async_verbose_true(get_email, mocker):
    """
    GIVEN a valid Email object
    WHEN Email.send_async() is called
    THEN assert it is added to the event loop for async sending
    """
    msg_loop_mock = mocker.patch.object(MESSAGELOOP, 'add_message')
    e = get_email
    e.verbose = True
    e.send_async()
    assert msg_loop_mock.call_count == 1



"""messages._config tests."""

import pytest
import builtins

from collections import OrderedDict

import jsonconfig

import messages._config
from messages._config import check_config_file
from messages._config import verify_profile_name
from messages._config import retrieve_data_from_config
from messages._config import retrieve_pwd_from_config
from messages._config import update_config_data
from messages._config import update_config_pwd
from messages._config import create_config_profile
from messages._config import display_required_items
from messages._config import get_user_ack
from messages._config import get_data_from_user
from messages._config import get_auth_from_user
from messages._config import configure_profile
from messages._config import write_data
from messages._config import write_auth
from messages._exceptions import UnknownProfileError
from messages._utils import credential_property


##############################################################################
# FIXTURES
##############################################################################

class Msg:
    """A test message class."""
    auth = credential_property('auth')
    def __init__(self, from_=None, auth=None, profile=None, save=False):
        self.from_ = from_
        self.auth = auth
        self.profile = profile
        self.save = save
    def __iter__(self):
        return iter(self.__dict__)


class Cfg:
    """basic cfg class."""
    def __init__(self):
        self.data = {}
        self.pwd = {}
        self.filename = '/default/path'


@pytest.fixture()
def get_msg():
    return Msg(from_='me', auth='passw0rd', profile='myProfile')


@pytest.fixture()
def get_cfg():
    return Cfg()

##############################################################################
# TESTS: check_config_file
##############################################################################

def test_check_config_file_auth_given_save_false(cfg_mock, get_msg, mocker):
    """
    GIVEN a call to check_config_file()
    WHEN auth is given and save=False
    THEN assert correct funcs are called
    """
    vrfy_mock = mocker.patch.object(messages._config, 'verify_profile_name')
    rdata_mock = mocker.patch.object(messages._config, 'retrieve_data_from_config')
    rpwd_mock = mocker.patch.object(messages._config, 'retrieve_pwd_from_config')
    udata_mock = mocker.patch.object(messages._config, 'update_config_data')
    upwd_mock = mocker.patch.object(messages._config, 'update_config_pwd')
    msg = get_msg
    check_config_file(msg)
    assert vrfy_mock.call_count == 1
    assert rdata_mock.call_count == 1
    assert rpwd_mock.call_count == 0
    assert udata_mock.call_count == 0
    assert upwd_mock.call_count == 0


def test_check_config_file_auth_Notgiven_save_false(cfg_mock, get_msg, mocker):
    """
    GIVEN a call to check_config_file()
    WHEN auth is not given and save=False
    THEN assert correct funcs are called
    """
    vrfy_mock = mocker.patch.object(messages._config, 'verify_profile_name')
    rdata_mock = mocker.patch.object(messages._config, 'retrieve_data_from_config')
    rpwd_mock = mocker.patch.object(messages._config, 'retrieve_pwd_from_config')
    udata_mock = mocker.patch.object(messages._config, 'update_config_data')
    upwd_mock = mocker.patch.object(messages._config, 'update_config_pwd')
    msg = get_msg
    msg.auth = None
    check_config_file(msg)
    assert vrfy_mock.call_count == 1
    assert rdata_mock.call_count == 1
    assert rpwd_mock.call_count == 1
    assert udata_mock.call_count == 0
    assert upwd_mock.call_count == 0


def test_check_config_file_auth_given_save_true(cfg_mock, get_msg, mocker):
    """
    GIVEN a call to check_config_file()
    WHEN auth is given and save=True
    THEN assert correct funcs are called
    """
    vrfy_mock = mocker.patch.object(messages._config, 'verify_profile_name')
    rdata_mock = mocker.patch.object(messages._config, 'retrieve_data_from_config')
    rpwd_mock = mocker.patch.object(messages._config, 'retrieve_pwd_from_config')
    udata_mock = mocker.patch.object(messages._config, 'update_config_data')
    upwd_mock = mocker.patch.object(messages._config, 'update_config_pwd')
    msg = get_msg
    msg.save = True
    check_config_file(msg)
    assert vrfy_mock.call_count == 1
    assert rdata_mock.call_count == 1
    assert rpwd_mock.call_count == 0
    assert udata_mock.call_count == 1
    assert upwd_mock.call_count == 1


##############################################################################
# TESTS: verify_profile_name
##############################################################################

def test_verify_profile_name_normal(get_msg, get_cfg):
    """
    GIVEN a call to verify_profile_name
    WHEN profile name is in cfg.data
    THEN assert no exeception is raised
    """
    msg = get_msg
    cfg = get_cfg
    cfg.data['myProfile'] = {}
    verify_profile_name(msg, cfg)


def test_verify_profile_name_raises(get_msg, get_cfg):
    """
    GIVEN a call to verify_profile_name
    WHEN profile name is not in cfg.data
    THEN assert UnknownProfileError is raised
    """
    msg = get_msg
    cfg = get_cfg
    with pytest.raises(UnknownProfileError):
        verify_profile_name(msg, cfg)


##############################################################################
# TESTS: retrieve_data_from_config
##############################################################################

def test_ret_data_from_config(get_msg, get_cfg):
    """
    GIVEN a call to retrive_data_from_config
    WHEN msg has attrs that are None and the same attrs have
        values stored in cfg
    THEN msg attrs are updated with cfg values
    """
    msg = get_msg
    cfg = get_cfg
    # create some attributes to test
    msg.a, msg.b = None, None
    cfg.data = {'myProfile': {'msg': {'a': 1, 'b': 2}}}
    retrieve_data_from_config(msg, cfg)
    assert msg.a == 1
    assert msg.b == 2


##############################################################################
# TESTS: retrieve_pwd_from_config
##############################################################################

def test_ret_pwd_from_config_singular(get_msg, get_cfg):
    """
    GIVEN a call to retrieve_pwd_from_config
    WHEN the pwd is saved in cfg.pwd and is a string without
     '  :  ' joining multiple strings
    THEN assert it is returned as a string
    """
    msg = get_msg
    cfg = get_cfg
    cfg.pwd = {'myProfile_msg': 'n3w_passw0rd'}
    retrieve_pwd_from_config(msg, cfg)
    assert msg._auth == 'n3w_passw0rd'


def test_ret_pwd_from_config_parsed(get_msg, get_cfg):
    """
    GIVEN a call to retrieve_pwd_from_config
    WHEN the pwd is saved in cfg.pwd and is a string with
     '  :  ' joining multiple strings
    THEN assert it is returned as a tuple split by the '  :  '
        joiner
    """
    msg = get_msg
    cfg = get_cfg
    cfg.pwd = {'myProfile_msg': 'n3w :: passw0rd'}
    retrieve_pwd_from_config(msg, cfg)
    assert msg._auth == ('n3w', 'passw0rd')


##############################################################################
# TESTS: update_config_data
##############################################################################

def test_update_cfg_data(get_msg, get_cfg):
    """
    GIVEN a call to update_config_data()
    WHEN msg has updated values
    THEN assert they are overwritten in cfg
    """
    msg = get_msg
    cfg = get_cfg
    # create some differing values to check
    msg.a, msg.b = 1, 2
    cfg.data = {'myProfile': {'a': 5, 'b': 6}}
    update_config_data(msg, cfg)
    assert cfg.data['myProfile']['a'] == 1
    assert cfg.data['myProfile']['b'] == 2


##############################################################################
# TESTS: update_config_pwd
##############################################################################

def test_update_cfg_pwd_singular(get_msg, get_cfg):
    """
    GIVEN a call to update_cfg_pwd
    WHEN msg.auth is a single string
    THEN assert it stored in cfg as such
    """
    msg = get_msg
    cfg = get_cfg
    msg.auth = 'n3w_password'
    cfg.pwd = {'myProfile_msg': 'passw0rd'}
    update_config_pwd(msg, cfg)
    assert cfg.pwd['myProfile_msg'] == 'n3w_password'


def test_update_cfg_pwd_parsed(get_msg, get_cfg):
    """
    GIVEN a call to update_cfg_pwd
    WHEN msg.auth is a tuple
    THEN assert it stored in cfg as a string with ' :: ' as the
        joiner
    """
    msg = get_msg
    cfg = get_cfg
    msg.auth = ('n3w', 'super', 'password')
    cfg.pwd = {'myProfile_msg': 'passw0rd'}
    update_config_pwd(msg, cfg)
    assert cfg.pwd['myProfile_msg'] == 'n3w :: super :: password'

##############################################################################
# TESTS: create_config_profile
##############################################################################

def test_create_config_profile_noAck(mocker):
    """
    GIVEN a call to create_config_profile
    WHEN the user does not say 'y' or 'Y'
    THEN assert the correct sequence is called
    """
    display_mock = mocker.patch.object(messages._config, 'display_required_items')
    gack_mock = mocker.patch.object(messages._config, 'get_user_ack')
    gack_mock.return_value = False
    input_mock = mocker.patch.object(builtins, 'input')
    data_mock = mocker.patch.object(messages._config, 'get_data_from_user')
    auth_mock = mocker.patch.object(messages._config, 'get_auth_from_user')
    conf_mock = mocker.patch.object(messages._config, 'configure_profile')
    create_config_profile('msg')
    assert display_mock.call_count == 1
    assert gack_mock.call_count == 1
    assert input_mock.call_count == 0
    assert data_mock.call_count == 0
    assert auth_mock.call_count == 0
    assert conf_mock.call_count == 0


def test_create_config_profile_Ack(mocker):
    """
    GIVEN a call to create_config_profile
    WHEN the user does say 'y' or 'Y'
    THEN assert the correct sequence is called
    """
    display_mock = mocker.patch.object(messages._config, 'display_required_items')
    gack_mock = mocker.patch.object(messages._config, 'get_user_ack')
    gack_mock.return_value = True
    input_mock = mocker.patch.object(builtins, 'input')
    data_mock = mocker.patch.object(messages._config, 'get_data_from_user')
    auth_mock = mocker.patch.object(messages._config, 'get_auth_from_user')
    conf_mock = mocker.patch.object(messages._config, 'configure_profile')
    create_config_profile('msg')
    assert display_mock.call_count == 1
    assert gack_mock.call_count == 1
    assert input_mock.call_count == 1
    assert data_mock.call_count == 1
    assert auth_mock.call_count == 1
    assert conf_mock.call_count == 1


##############################################################################
# TESTS: display_required_items
##############################################################################

@pytest.mark.parametrize(['msg', 'expected'],[
    ('email', ['From email address', 'Email server url', 'Email server port number', 'Email service password']),
    ('slackpost', ['Name or alias of sender', 'Channel to post message to', 'Slack API authentication token']),
    ('slackwebhook', ['Name or alias of sender', 'Slack API Webhook URL']),
    ('telegrambot', ['Telegram Channel ID of chat', 'Telegram authorization token']),
    ('twilio', ['Twilio phone numbe', 'Twilio API account SID', 'Twilio API authorization token']),
    ('whatsapp', ['Twilio phone numbe', 'Twilio API account SID', 'Twilio API authorization token'])
])
def test_display_required_items(msg, expected, capsys):
    """
    GIVEN a call to display_required_items
    WHEN given a message type
    THEN assert the correct output prints
    """
    display_required_items(msg)
    out, err = capsys.readouterr()
    assert 'Configure a profile for:' in out
    assert 'You will need the following information:' in out
    assert 'Authorization/credentials required:' in out
    for e in expected:
        assert e in out


##############################################################################
# TESTS: get_user_ack
##############################################################################

def test_get_user_ack_yes(mocker):
    """
    GIVEN a call to get_user_ack()
    WHEN the user gives 'y'
    THEN assert True is returned
    """
    input_mock = mocker.patch.object(builtins, 'input')
    input_mock.return_value = 'y'
    ack = get_user_ack()
    assert ack == True


def test_get_user_ack_no(mocker):
    """
    GIVEN a call to get_user_ack()
    WHEN the user gives 'n'
    THEN assert False is returned
    """
    input_mock = mocker.patch.object(builtins, 'input')
    input_mock.return_value = 'n'
    ack = get_user_ack()
    assert ack == False

##############################################################################
# TESTS: get_data_from_user
##############################################################################

@pytest.mark.parametrize(['msg', 'expected'], [
    ('email', 3),
    ('slackpost', 2),
    ('slackwebhook', 1),
    ('telegrambot', 1),
    ('twilio', 1),
    ('whatsapp', 1)
])
def test_get_data_from_user(msg, expected, mocker):
    """
    GIVEN a call to get_data_from_user
    WHEN a specific message is specified
    THEN assert input is called the correct number of times
    """
    input_mock = mocker.patch.object(builtins, 'input')
    get_data_from_user(msg)
    assert input_mock.call_count == expected


##############################################################################
# TESTS: get_auth_from_user
##############################################################################

@pytest.mark.parametrize(['msg', 'expected'], [
    ('email', 1),
    ('slackpost', 1),
    ('slackwebhook', 1),
    ('telegrambot', 1),
    ('twilio', 2),
    ('whatsapp', 2)
])
def test_get_auth_from_user(msg, expected, mocker):
    """
    GIVEN a call to get_auth_from_user
    WHEN a specific message is specified
    THEN assert getpass is called the correct number of times
    """
    getpass_mock = mocker.patch.object(messages._config, 'getpass')
    get_auth_from_user(msg)
    assert getpass_mock.call_count == expected


##############################################################################
# TESTS: configure_profile
##############################################################################


def test_configure_profile(cfg_mock, mocker, capsys):
    """
    GIVEN a call to configure_profile()
    WHEN given valid arguments
    THEN assert the correct funcs called and correct output prints
    """
    data_mock = mocker.patch.object(messages._config, 'write_data')
    auth_mock = mocker.patch.object(messages._config, 'write_auth')
    profile_name = 'newProfile'
    data = {'a': 1, 'b': 2}
    auth = {'auth': 'passw0rd'}
    configure_profile('msg', profile_name, data, auth)
    out, err = capsys.readouterr()
    assert data_mock.call_count == 1
    assert auth_mock.call_count == 1
    assert '[+] Configuration entry for <' in out
    assert profile_name in out
    assert '> created.' in out
    assert '[+] Configuration file location:' in out
    assert '/default/path' in out


##############################################################################
# TESTS: write_data
##############################################################################

def test_write_data(get_cfg):
    """
    GIVEN a call to write_data
    WHEN data params are given
    THEN assert the params are correctly stored
    """
    cfg = get_cfg
    profile_name = 'newProfile'
    data = {'a': 1, 'b': 2}
    msg_type = 'msg'
    write_data(msg_type, profile_name, data, cfg)
    assert cfg.data['newProfile']['msg'] == data


##############################################################################
# TESTS: write_auth
##############################################################################

@pytest.mark.parametrize(['msg', 'auth', 'pwd'], [
    ('email', {'auth': 's3cr3t'}, 's3cr3t'),
    ('slackpost', {'auth': 's3cr3t'}, 's3cr3t'),
    ('slackwebhook', {'auth': 's3cr3t'}, 's3cr3t'),
    ('telegrambot', {'auth': 's3cr3t'}, 's3cr3t'),
    ('twilio', OrderedDict([('auth_sid', 'ABCD'), ('auth_token', '1234')]),
        'ABCD :: 1234'),
    ('whatsapp', OrderedDict([('auth_sid', 'ABCD'), ('auth_token', '1234')]),
        'ABCD :: 1234')
])
def test_write_auth(msg, auth, pwd, get_cfg):
    """
    GIVEN a call to write_auth
    WHEN an auth param is given
    THEN assert it is correctly stored
    """
    cfg = get_cfg
    profile_name = 'newProfile'
    key_fmt = profile_name + '_' + msg
    write_auth(msg, profile_name, auth, cfg)
    assert cfg.pwd[key_fmt] == pwd



"""Utility Module - functions useful to other modules."""

import datetime
from collections import MutableSequence

import validus

from ._exceptions import InvalidMessageInputError


"""
Functions below this header are property factories and input validation functions
used by each of the message classes in order to obscure (credentials) or validate
certain attributes.  Each will be defined as a class attribute for each message
(before __init__).
"""


def credential_property(cred):
    """
    A credential property factory for each message class that will set
    private attributes and return obfuscated credentials when requested.
    """

    def getter(instance):
        return "***obfuscated***"

    def setter(instance, value):
        private = "_" + cred
        instance.__dict__[private] = value

    return property(fget=getter, fset=setter)


def validate_property(attr):
    """
    A property factory that will dispatch the to a specific validator function
    that will validate the user's input to ensure critical parameters are of a
    specific type.
    """

    def getter(instance):
        return instance.__dict__[attr]

    def setter(instance, value):
        validate_input(instance.__class__.__name__, attr, value)
        instance.__dict__[attr] = value

    return property(fget=getter, fset=setter)


def validate_input(msg_type, attr, value):
    """Base function to validate input, dispatched via message type."""
    try:
        valid = {
            "Email": validate_email,
            "Twilio": validate_twilio,
            "SlackWebhook": validate_slackwebhook,
            "SlackPost": validate_slackpost,
            "TelegramBot": validate_telegrambot,
            "WhatsApp": validate_whatsapp,
        }[msg_type](attr, value)
    except KeyError:
        return 1
    else:
        return 0


def check_valid(msg_type, attr, value, func, exec_info):
    """
    Checker function all validate_* functions below will call.
    Raises InvalidMessageInputError if input is not valid as per
    given func.
    """
    if value is not None:
        if isinstance(value, MutableSequence):
            for v in value:
                if not func(v):
                    raise InvalidMessageInputError(msg_type, attr, value, exec_info)
        else:
            if not func(value):
                raise InvalidMessageInputError(msg_type, attr, value, exec_info)


def validate_email(attr, value):
    """Email input validator function."""
    check_valid("Email", attr, value, validus.isemail, "email address")


def validate_twilio(attr, value):
    """Twilio input validator function."""
    if attr in ("from_", "to"):
        check_valid("Twilio", attr, value, validus.isphone, "phone number")
    elif attr in ("attachments"):
        check_valid("Twilio", attr, value, validus.isurl, "url")


def validate_slackwebhook(attr, value):
    """SlackWebhook input validator function."""
    check_valid("SlackWebhook", attr, value, validus.isurl, "url")


def validate_slackpost(attr, value):
    """SlackPost input validator function."""
    if attr in ("channel", "credentials"):
        if not isinstance(value, str):
            raise InvalidMessageInputError("SlackPost", attr, value, "string")
    elif attr in ("attachments"):
        check_valid("SlackPost", attr, value, validus.isurl, "url")


def validate_telegrambot(attr, value):
    """TelegramBot input validator function."""
    check_valid("TelegramBot", attr, value, validus.isint, "integer as a string")


def validate_whatsapp(attr, value):
    """WhatsApp input validator function."""
    if attr in ("from_", "to"):
        value = value.split("whatsapp:+")[-1]
        check_valid(
            "WhatsApp",
            attr,
            value,
            validus.isint,
            "phone number starting with the '+' symbol",
        )
    elif attr in ("attachments"):
        check_valid("WhatsApp", attr, value, validus.isurl, "url")


"""
Functions below this hearder are general utility functions.
"""


def timestamp():
    """Get current date and time."""
    return "{:%Y-%b-%d %H:%M:%S}".format(datetime.datetime.now())



"""
messages - A package designed to make sending various types of messages easy.
"""

import logging

from .api import send

from .email_ import Email
from .slack import SlackWebhook
from .slack import SlackPost
from .telegram import TelegramBot
from .text import Twilio
from .whatsapp import WhatsApp


__version__ = "0.5.0"


# Setup logger
logging.getLogger(__name__).addHandler(logging.NullHandler())



"""messages._utils tests."""

import builtins
import re

import pytest

import messages._utils
from messages._utils import credential_property
from messages._utils import validate_property
from messages._utils import validate_input
from messages._utils import check_valid
from messages._utils import validate_email
from messages._utils import validate_twilio
from messages._utils import validate_slackwebhook
from messages._utils import validate_slackpost
from messages._utils import validate_telegrambot
from messages._utils import validate_whatsapp
from messages._utils import timestamp
from messages._utils import validus
from messages._exceptions import InvalidMessageInputError


##############################################################################
# FIXTURES
##############################################################################

class DummyClass:
    """Test class used for testing."""
    cred_param = credential_property('cred_param')
    normal_param = validate_property('normal_param')
    def __init__(self, cred_param, normal_param):
        self.cred_param, self.normal_param = cred_param, normal_param


@pytest.fixture()
def get_test_class():
    """Return an instance of TestClass."""
    return DummyClass('s3cr3t', 'information')


def val_test_func(item):
    """Test func for check_valid."""
    if item == 'BAD':
        return False
    return True


##############################################################################
# TEST: credential_property & validate_property
##############################################################################

def test_properties_instantiation(get_test_class, mocker):
    """
    GIVEN a message class that utilized both *_property factory functions
    WHEN instantiation occurs
    THEN assert credential_property and validate_property successfully execute
    """
    val_mock = mocker.patch.object(messages._utils, 'validate_input')
    t = get_test_class
    assert t.cred_param == '***obfuscated***'
    assert '_cred_param' in t.__dict__
    assert t._cred_param == 's3cr3t'
    assert t.normal_param == 'information'
    t.normal_param = 'new information'
    t.normal_param = 'even newer information'
    assert val_mock.call_count == 2


def test_credential_property(mocker):
    """
    GIVEN an attribute to make as a property
    WHEN credential_property is called
    THEN assert a property is returned
    """
    val_mock = mocker.patch.object(messages._utils, 'validate_input')
    param = credential_property('param')
    assert isinstance(param, property)


def test_validate_property(mocker):
    """
    GIVEN an attribute to make as a property
    WHEN validate_property is called
    THEN assert a property is returned
    """
    val_mock = mocker.patch.object(messages._utils, 'validate_input')
    param = validate_property('param')
    assert isinstance(param, property)


##############################################################################
# TEST: validate_input
##############################################################################

def test_val_input_NotSupported():
    """
    GIVEN a message object is instantiated
    WHEN validate_input() is called on a message object that is
        not supported for input validation
    THEN assert nothing happens
    """
    result = validate_input('NotSupported', 'attr', 'some_value')
    assert result == 1


@pytest.mark.parametrize('msg_type, func, result', [
    ('Email', 'validate_email', 0),
    ('Twilio', 'validate_twilio', 0),
    ('SlackPost', 'validate_slackpost', 0),
    ('SlackWebhook', 'validate_slackwebhook', 0),
    ('TelegramBot', 'validate_telegrambot', 0),
    ('WhatsApp', 'validate_whatsapp', 0),
])
def test_val_input_supported(msg_type, func, result, mocker):
    """
    GIVEN a message object is instantiated
    WHEN validate_input() is called on a message object
    THEN assert the proper validate_* func is called and 0 is the return value
    """
    val_mock = mocker.patch.object(messages._utils, func)
    r = validate_input(msg_type, 'attr', 'some_value')
    assert val_mock.call_count == 1
    assert r == result


##############################################################################
# TEST: validate_*
##############################################################################

@pytest.mark.parametrize('msg_type, func, attr', [
    ('Email', validate_email, 'address'),
    ('Twilio', validate_twilio, 'from_'),
    ('Twilio', validate_twilio, 'attachments'),
    ('SlackWebhook', validate_slackwebhook, 'url'),
    ('TelegramBot', validate_telegrambot, 'id_num' ),
    ('WhatsApp', validate_whatsapp, 'from_'),
    ('WhatsApp', validate_whatsapp, 'attachments')
])
def test_val_funcs(msg_type, func, attr, mocker):
    """
    GIVEN a message instance
    WHEN validate_* is called
    THEN assert check_valid is called the requisite number of times
    """
    check_mock = mocker.patch.object(messages._utils, 'check_valid')
    func(attr, 'some_value')
    assert check_mock.call_count == 1


def test_val_slackPost(mocker):
    """
    GIVEN a slackPost instance
    WHEN validate_slackpost is called with the correct input type
    THEN assert nothing bad happens
    **SlackPost is tested separately from the parametrized test_val_func()
    test because validate_slackpost is written differently and cannot be
    tested the same way.
    """
    check_mock = mocker.patch.object(messages._utils, 'check_valid')
    validate_slackpost('channel', 'some_value')
    validate_slackpost('attachments', 'some_value')


def test_val_slackPost_raises():
    """
    GIVEN a slackPost instance
    WHEN validate_slackpost is called with incorrect input type
    THEN assert InvalideMessageInputError is raised
    **SlackPost is tested separately from the parametrized test_val_func()
    test because validate_slackpost is written differently and cannot be
    tested the same way.
    """
    with pytest.raises(InvalidMessageInputError):
        validate_slackpost('channel', 1)


##############################################################################
# TEST: check_valid
##############################################################################

def test_check_valid(get_test_class):
    """
    GIVEN a valid message object
    WHEN check_valid is called on requisite attributes
    THEN assert normal behavior and no exceptions raised
    """
    t = get_test_class
    for key, value in t.__dict__.items():
        check_valid('TestClass', key, value, val_test_func, 'required type')


def test_check_valid_listAttributes(get_test_class):
    """
    GIVEN a valid message object
    WHEN check_valid is called on requisite attributes that are lists
    THEN assert normal behavior and no exceptions raised
    """
    t = get_test_class
    t.listAttr = ['val1', 'val2']
    for key, value in t.__dict__.items():
        check_valid('TestClass', key, value, val_test_func, 'required type')


def test_check_valid_singleton_raisesExc(get_test_class):
    """
    GIVEN a message object with a single invalid input
    WHEN check_valid is called
    THEN assert InvalidMessageInputError is raised
    """
    t = get_test_class
    t.normal_param = 'BAD'
    with pytest.raises(InvalidMessageInputError):
        for key, value in t.__dict__.items():
            check_valid('TestClass', key, value, val_test_func, 'required type')


def test_check_valid_list_raisesExc(get_test_class):
    """
    GIVEN a message object with a list of invalid inputs
    WHEN check_valid is called
    THEN assert InvalidMessageInputError is raised
    """
    t = get_test_class
    t.normal_param = ['GOOD', 'BAD']
    with pytest.raises(InvalidMessageInputError):
        for key, value in t.__dict__.items():
            check_valid('TestClass', key, value, val_test_func, 'required_type')


##############################################################################
# TEST: timestamp
##############################################################################

def test_timestamp():
    """
    GIVEN a need for a timestamp (i.e. --verbose output)
    WHEN timestamp() is called
    THEN assert a timestamp as a string is returned
    """
    t = timestamp()
    #r = '^\d{4}-[A-Za-z]{3}-\d{1,2}\s{1}\d{2}:\d{2}:\d{2}$'
    #assert re.match(r, t)
    assert isinstance(t, str)



"""messages.slack tests."""

import pytest
import requests

import messages.telegram
from messages.telegram import TelegramBot
from messages._eventloop import MESSAGELOOP


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_tgram(cfg_mock, mocker):
    """Return a valid TelegramBot object."""
    mocker.patch.object(messages.telegram, 'check_config_file')
    return TelegramBot(auth='34563:ABCDEFG', chat_id='123456', body='message',
                attachments=['https://url1.com', 'https://url2.com'],
                profile='myProfile')


##############################################################################
# TESTS: TelegramBot.__init__
##############################################################################

def test_tgram_init(get_tgram):
    """
    GIVEN a need for a TelegramBot object
    WHEN instantiated
    THEN assert it is properly created
    """
    t = get_tgram
    assert t.body == 'message'
    assert t.auth == '***obfuscated***'
    assert '_auth' in t.__dict__
    assert t._auth == '34563:ABCDEFG'
    assert isinstance(t.message, dict)


##############################################################################
# TESTS: TelegramBot.__str__
##############################################################################

def test_tgram_str(get_tgram, capsys):
    """
    GIVEN t=TelegramBot()
    WHEN print(t)
    THEN assert proper output prints
    """
    t = get_tgram
    t.from_ = 'Me'
    t.to = 'You'
    t.subject = 'Chat'
    print(t)
    out, err = capsys.readouterr()
    assert 'From: Me' in out
    assert 'To: You' in out
    assert 'Chat ID: 123456' in out
    assert 'Subject: Chat' in out
    assert 'Body: \'message\'' in out
    assert 'Attachments: [\'https://url1.com\', \'https://url2.com\']' in out
    assert '' in err


##############################################################################
# TESTS: TelegramBot.get_chat_id
##############################################################################

def test_tgram_getChatID(get_tgram, mocker):
    """
    GIVEN a TelegramBot instance with unknown chat_id of recipient
    WHEN get_chat_id() is called
    THEN assert proper data is returned
    """
    req_mock = mocker.patch.object(requests, 'get')
    req_mock.return_value.json.return_value = {'result': [{'message':{
        'from':{'username': 'YOU', 'id': '123456'}}}]}
    t = get_tgram
    id = t.get_chat_id('@YOU')
    assert id == '123456'


##############################################################################
# TESTS: TelegramBot._construct_message
##############################################################################

def test_tgram_construct_message(get_tgram):
    """
    GIVEN a valid TelegramBot object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    t = get_tgram
    t._construct_message()
    assert t.message['text'] == 'message'


def test_tgram_construct_message_withFromSub(get_tgram):
    """
    GIVEN a valid TelegramBot object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    t = get_tgram
    t.from_ = 'me'
    t.subject = 'this is a test'
    t._construct_message()
    expected = 'From: me\n' + 'Subject: this is a test\n' + 'message'
    assert t.message['text'] == expected


##############################################################################
# TESTS: TelegramBot._send_content
##############################################################################

def test_tgram_send_content_verbose_false(get_tgram, capsys, mocker):
    """
    GIVEN a valid TelegramBot object
    WHEN send_content() is called with verbose=False
    THEN assert the proper send sequence occurs
    """
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 200
    t = get_tgram
    t._send_content()
    out, err = capsys.readouterr()
    assert req_mock.call_count == 1
    assert out == ''
    assert err == ''


def test_tgram_send_content_msgBody_verbose_true(get_tgram, capsys, mocker):
    """
    GIVEN a valid TelegramBot object
    WHEN send_content() is called with verbose=True
    THEN assert the proper send sequence occurs
    """
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 200
    t = get_tgram
    t.verbose = True
    t._send_content()
    out, err = capsys.readouterr()
    assert req_mock.call_count == 1
    assert 'Message body sent' in out
    assert 'Attachment: https://url1.com' not in out
    assert 'Attachment: https://url2.com' not in out
    assert err == ''


def test_tgram_send_content_attachments_verbose_true(get_tgram, capsys, mocker):
    """
    GIVEN a valid TelegramBot object
    WHEN send_content() is called with verbose=True
    THEN assert the proper send sequence occurs
    """
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 200
    t = get_tgram
    t.verbose = True
    t.message['document'] = 'https://url1.com'
    t._send_content(method='/sendDocument')
    out, err = capsys.readouterr()
    assert req_mock.call_count == 1
    assert 'Message body sent' not in out
    assert 'Attachment: https://url1.com' in out
    assert err == ''


def test_tgram_send_content_statusGT300_verbose_true(get_tgram, capsys, mocker):
    """
    GIVEN a valid TelegramBot object
    WHEN send_content() is called but an error occurs, status_code > 300, verbose=True
    THEN assert the proper send sequence occurs
    """
    req_mock = mocker.patch.object(requests, 'post')
    req_mock.return_value.status_code = 403
    req_mock.return_value.text = 'test error'
    t = get_tgram
    t.verbose = True
    t._send_content()
    out, err = capsys.readouterr()
    assert req_mock.call_count == 1
    assert 'Error while sending.  HTTP status code: 403' in out
    assert 'test error' in out
    assert err == ''

##############################################################################
# TESTS: TelegramBot.send
##############################################################################

def test_send_verbose_false(get_tgram, mocker, capsys):
    """
    GIVEN a TelegramBot instance
    WHEN send() is called with verbose=False
    THEN assert correct sequence is called and correct output printed
    """
    con_mock = mocker.patch.object(TelegramBot, '_construct_message')
    send_cont_mock = mocker.patch.object(TelegramBot, '_send_content')
    t = get_tgram
    t.send()
    out, err = capsys.readouterr()
    assert con_mock.call_count == 1
    assert send_cont_mock.call_count == 3
    assert 'Message sent.' in out
    assert 'Debugging info' not in out


def test_send_verbose_true(get_tgram, mocker, capsys):
    """
    GIVEN a TelegramBot instance
    WHEN send() is called with verbose=True
    THEN assert correct sequence is called and correct output printed
    """
    con_mock = mocker.patch.object(TelegramBot, '_construct_message')
    send_cont_mock = mocker.patch.object(TelegramBot, '_send_content')
    t = get_tgram
    t.verbose = True
    t.send()
    out, err = capsys.readouterr()
    assert con_mock.call_count == 1
    assert send_cont_mock.call_count == 3
    assert 'Message sent.' in out
    assert 'Debugging info' in out


##############################################################################
# TESTS: TelegramBot.send_async
##############################################################################

def test_tgram_send_async(get_tgram, mocker):
    """
    GIVEN a valid TelegramBot object
    WHEN send_async() is called
    THEN assert the proper send sequence occurs
    """
    msgloop_mock = mocker.patch.object(MESSAGELOOP, 'add_message')
    t = get_tgram
    t.send_async()
    assert msgloop_mock.call_count == 1



"""
Module designed to make creating and sending WhatsApp messages easy.

1.  WhatsApp
    - Uses the Twilio API to send WhatsApp messages, thus inherits from the
      Twilio class.
    - Must have an account_sid, auth_token, and a twilio phone number
      in order to use.
    - Go to https://www.twilio.com/whatsapp to register.
"""

from ._config import check_config_file
from ._utils import credential_property
from ._utils import validate_property
from .text import Twilio


class WhatsApp(Twilio):
    """
    Create and send WhatsApp messages using the WhatsApp API provided by
    Twilio.

    Inheritance:
        :Twilio: Inherits functionality from Twilio class.  See the
            messages.text.Twilio class.

    Args:
        :from_: (str) phone number of originating text, e.g. '+15558675309'
        :to: (str) phone number of destination text, e.g. '+15558675309'
        :auth: (list or tuple) twilio api credentials: (acct_sid, auth_token)
        :body: (str) message to send
        :attachments: (str) url of any image to send along with message
        :profile: (str) use a separate account profile specified by name
        :save: (bool) save pertinent values in the messages config file,
            such as from_, acct_sid, auth_token (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :sid: (str) return value from send, record of sent message

    Managed Attributes (Properties):
        :auth: auth will set as a private attribute (_auth) and obscured when requested
        :from_: user input will be validated for a proper phone number
        :to: user input will be validated for a proper phone number
        :attachments: user input will be validated for a proper url

    Usage:
        Create a WhatsApp message instance with required Args above.
        Send message with self.send() or self.send_async() methods.

    Notes:
        For API description:
        https://www.twilio.com/docs/sms/whatsapp/quickstart
    """

    auth = credential_property("auth")
    from_ = validate_property("from_")
    to = validate_property("to")
    attachments = validate_property("attachments")

    def __init__(
        self,
        from_=None,
        to=None,
        auth=None,
        body="",
        attachments=None,
        profile=None,
        save=False,
        verbose=False,
    ):

        self.from_ = "whatsapp:" + from_
        self.to = "whatsapp:" + to
        self.auth = auth
        self.body = body
        self.attachments = attachments
        self.profile = profile
        self.save = save
        self.verbose = verbose
        self.sid = None

        if self.profile:
            check_config_file(self)



"""messages._interface tests."""

import pytest

import messages._interface
from messages._interface import Message


##############################################################################
# FIXTURES
##############################################################################

class MsgGood(Message):
    """Good Message class since it has all required abstract methods."""
    def __init__(self, x, y):
        self.x, self.y = x, y
    def send(self): pass
    def send_async(self): pass


class MsgBad1(Message):
    """Bad Message class since it is missing one required abstract methods."""
    def __init__(self): pass
    def send(self): pass


class MsgBad2(Message):
    """Bad Message class since it is missing one required abstract methods."""
    def __init__(self): pass
    def send_async(self): pass


##############################################################################
# TESTS: instantiation
##############################################################################

def test_MsgGood():
    """
    GIVEN a message class that inherits from 'Message'
    WHEN instantiated with all required abstract methods
    THEN assert it is created
    """
    msg = MsgGood(1, 2)
    assert msg is not None


def test_MsgBad1():
    """
    GIVEN a message class that inherits from 'Message'
    WHEN instantiated with missing required abstract methods
    THEN assert
    """
    with pytest.raises(TypeError):
        msg = MsgBad1()


def test_MsgBad2():
    """
    GIVEN a message class that inherits from 'Message'
    WHEN instantiated with missing required abstract methods
    THEN assert
    """
    with pytest.raises(TypeError):
        msg = MsgBad2()


##############################################################################
# TESTS: __repr__
##############################################################################

def test_repr(capsys):
    """
    GIVEN a message class that inherits from 'Message'
    WHEN repr(msg) or `>>> msg` is called
    THEN assert appropriate output is printed
    """
    msg = MsgGood(1, 2)
    msg._auth = 'password'
    msg.body = 'A'*50
    print(repr(msg))
    out, err = capsys.readouterr()
    assert 'MsgGood(' in out
    #assert 'auth=***obfuscated***,' in out
    assert 'body=' in out
    assert 'A...A' in out
    assert 'x=1,' in out
    assert 'y=2' in out
    assert ')' in out
    assert err == ''


##############################################################################
# TESTS: __iter__
##############################################################################

def test_iter():
    """
    GIVEN a message class that inherits from 'Message'
    WHEN an iterator-type function is called (such as set(msg))
    THEN assert all defined attributes are returned
    """
    msg = MsgGood(1, 2)
    s = set(msg)
    assert s == {'x', 'y'}



"""messages.api tests."""

import pytest

import messages.api
from messages.api import send
from messages.api import message_factory
from messages.email_ import Email
from messages.slack import SlackWebhook
from messages.slack import SlackPost
from messages.telegram import TelegramBot
from messages.text import Twilio
from messages._exceptions import UnsupportedMessageTypeError


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def email_kwargs():
    return {'server': 'smtp.gmail.com', 'port': 465,
            'auth': 'passw0rd', 'from_': 'me@here.com',
            'to': 'you@there.com', 'cc': None, 'bcc': None,
            'subject': 'TEST', 'body': 'this is a message',
            'attachments': None}


@pytest.fixture()
def slackwebhook_kwargs():
    return {'auth': 'https://slack.com', 'body': 'Test message',
            'attachments': None, 'params': {'author_name': 'me'}}


@pytest.fixture()
def slackpost_kwargs():
    return {'auth': '12345abcdef', 'channel': 'general',
            'body': 'Test message', 'attachments': None,
            'params': {'author_name': 'me'}}


@pytest.fixture()
def telegrambot_kwargs():
    return {'auth': '1234:ABCD', 'chat_id': '123456'}


@pytest.fixture()
def twilio_kwargs():
    return {'auth': ('your sid', 'your token'),
            'from_': '+19998675309', 'to': '+19998675309', 'body': 'Test!',
            'attachments': 'https://www.google.com'}


##############################################################################
# TESTS: send()
##############################################################################

def test_send_async_false(email_kwargs, mocker):
    """
    GIVEN a need to create and send an email message
    WHEN api.send() is called
    THEN assert first the factory function returns a valid message instance
    """
    kwargs = email_kwargs
    fact_mock = mocker.patch.object(messages.api, 'message_factory')
    send('email', **kwargs)
    assert fact_mock.call_count == 1


def test_send_async_true(email_kwargs, mocker):
    """
    GIVEN a need to create and send an email message
    WHEN api.send() is called
    THEN assert first the factory function returns a valid message instance
    """
    kwargs = email_kwargs
    fact_mock = mocker.patch.object(messages.api, 'message_factory')
    send('email', send_async=True, **kwargs)
    assert fact_mock.call_count == 1


##############################################################################
# TESTS: message_factory
##############################################################################

def test_message_factory_keyerror(email_kwargs, cfg_mock):
    """
    GIVEN a need to create a message object
    WHEN message_factory is called with an unsupported message type
    THEN assert UnsupportedMessageTypeError is raised
    """
    kwargs = email_kwargs
    with pytest.raises(UnsupportedMessageTypeError):
        msg = message_factory('bad', **kwargs)


def test_message_factory_email(email_kwargs, cfg_mock):
    """
    GIVEN a need to create an email message with the specified kwargs
    WHEN message_factory is called
    THEN assert an Email instance is returned
    """
    kwargs = email_kwargs
    msg = message_factory('email', **kwargs)
    assert isinstance(msg, Email)


def test_message_factory_slackwebhook(slackwebhook_kwargs, cfg_mock):
    """
    GIVEN a need to create a slackwebhook message with the specified kwargs
    WHEN message_factory is called
    THEN assert a SlackWebhook instance is returned
    """
    kwargs = slackwebhook_kwargs
    msg = message_factory('slackwebhook', **kwargs)
    assert isinstance(msg, SlackWebhook)


def test_message_factory_slackpost(slackpost_kwargs, cfg_mock):
    """
    GIVEN a need to create a slackpost message with the specified kwargs
    WHEN message_factory is called
    THEN assert a SlackPost instance is returned
    """
    kwargs = slackpost_kwargs
    msg = message_factory('slackpost', **kwargs)
    assert isinstance(msg, SlackPost)


def test_message_factory_telegrambot(telegrambot_kwargs, cfg_mock):
    """
    GIVEN a need to create a telegrambot message with the specified kwargs
    WHEN message_factory is called
    THEN assert a TelegramBot instance is returned
    """
    kwargs = telegrambot_kwargs
    msg = message_factory('telegrambot', **kwargs)
    assert isinstance(msg, TelegramBot)


def test_message_factory_twilio(twilio_kwargs, cfg_mock):
    """
    GIVEN a need to create a twilio message with the specified kwargs
    WHEN message_factory is called
    THEN assert a Twilio instance is returned
    """
    kwargs = twilio_kwargs
    msg = message_factory('twilio', **kwargs)
    assert isinstance(msg, Twilio)



"""messages.text tests."""

import pytest
import requests

import messages.text
from messages.text import Twilio
from messages.text import check_config_file
from messages._eventloop import MESSAGELOOP


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_twilio(mocker):
    """Return a valid Twilio object."""
    configure_mock = mocker.patch.object(messages.text, 'check_config_file')
    t = Twilio(from_='+16198675309', to='+16195551212',
            auth=('test_sid', 'test_token'), body='test text!',
            attachments='https://imgs.xkcd.com/comics/python.png',
            profile='tester', save=False)
    t.from_ = '+16198675309'
    t.auth = ('test_sid', 'test_token')
    t.profile = 'tester'
    return t


##############################################################################
# TESTS: Twilio.__init__
##############################################################################

def test_twilio_init(get_twilio, cfg_mock):
    """
    GIVEN a need to create an Twilio object
    WHEN the user instantiates a new object with required args
    THEN assert Twilio object is created with given args
    """
    t = get_twilio
    assert t.from_ == '+16198675309'
    assert t.to == '+16195551212'
    assert t.auth == '***obfuscated***'
    assert '_auth' in t.__dict__
    assert t._auth == ('test_sid', 'test_token')
    assert t.body == 'test text!'
    assert t.attachments == 'https://imgs.xkcd.com/comics/python.png'


##############################################################################
# TESTS: Twilio.__str__
##############################################################################

def test_twilio_str(get_twilio,cfg_mock, capsys):
    """
    GIVEN a valid Twilio object
    WHEN the user calls print() on the Twilio object
    THEN assert the correct format prints
    """
    t = get_twilio
    expected = ('\nFrom: +16198675309'
                '\nTo: +16195551212'
                '\nBody: \'test text!\''
                '\nAttachments: https://imgs.xkcd.com/comics/python.png'
                '\nSID: None\n')
    print(t)
    out, err = capsys.readouterr()
    assert out == expected
    assert err == ''


##############################################################################
# TESTS: Twilio.send
##############################################################################

def test_send_verbose_false_status201(get_twilio, cfg_mock, capsys, mocker):
    """
    GIVEN a valid Twilio object
    WHEN Twilio.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to False)
    """
    msg_mock = mocker.patch.object(requests, 'post')
    msg_mock.return_value.status_code = 201
    t = get_twilio
    t.send()
    t.sid = 12345
    out, err = capsys.readouterr()
    assert 'Debugging info' not in out
    assert 'Message created.' not in out
    assert '* From: +16198675309' not in out
    assert '* To: +16195551212' not in out
    assert '* Body: test text!' not in out
    assert '* Attachments: https://imgs.xkcd.com/comics/python.png' not in out
    assert '* SID: ' not in out
    assert '* HTTP status code:' not in out
    assert out == 'Message sent.\n'
    assert err == ''


def test_send_verbose_true_status201(get_twilio, cfg_mock, capsys, mocker):
    """
    GIVEN a valid Twilio object
    WHEN Twilio.send() is called
    THEN assert the correct functions are called, correct attributes
        updated and correct debug output is generated (using verbose flag
        set to True)
    """
    msg_mock = mocker.patch.object(requests, 'post')
    msg_mock.return_value.status_code = 201
    t = get_twilio
    t.verbose = True
    t.send()
    out, err = capsys.readouterr()
    assert 'Debugging info' in out
    assert 'Message created.' in out
    assert '* From: +16198675309' in out
    assert '* To: +16195551212' in out
    assert '* Body: \'test text!\'' in out
    assert '* Attachments: https://imgs.xkcd.com/comics/python.png' in out
    assert '* SID: ' in out
    assert '* HTTP status code: 201' in out
    assert 'Message sent.' in out
    assert err == ''


def test_send_statusGT300(get_twilio, cfg_mock, capsys, mocker):
    """
    GIVEN a valid Twilio object
    WHEN Twilio.send() is called but an HTTP status code >= 300 is returned
    THEN assert 'Error while sending' prints
    """
    msg_mock = mocker.patch.object(requests, 'post')
    msg_mock.return_value.status_code = 301
    t = get_twilio
    t.send()
    out, err = capsys.readouterr()
    assert 'Error while sending.' in out
    assert err == ''


##############################################################################
# TESTS: Twilio.send_async
##############################################################################

def test_send_async(cfg_mock, get_twilio, mocker):
    """
    GIVEN a valid Twilio object
    WHEN Twilio.send_async() is called
    THEN assert it is added to the event loop for async sending
    """
    msg_loop_mock = mocker.patch.object(MESSAGELOOP, 'add_message')
    t = get_twilio
    t.send_async()
    assert msg_loop_mock.call_count == 1



"""
Module designed to make creating and sending text messages easy.

1.  Twilio
    - Uses the Twilio API to send text messages.
    - Must have an account_sid, auth_token, and a twilio phone number
      in order to use.
    - Go to https://www.twilio.com to register.
"""

import reprlib

import requests

from ._config import check_config_file
from ._eventloop import MESSAGELOOP
from ._interface import Message
from ._utils import credential_property
from ._utils import validate_property
from ._utils import timestamp


class Twilio(Message):
    """
    Create and send text SMS/MMS text messages using the Twilio API.

    Args:
        :from_: (str) phone number of originating text, e.g. '+15558675309'
        :to: (str) phone number of destination text, e.g. '+15558675309'
        :auth: (list or tuple) twilio api credentials: (acct_sid, auth_token)
        :body: (str) message to send.  Defaults to body=" " (one space) since
            Twilio doesn't allow an empty body to be sent
        :attachments: (str) url of any image to send along with message
        :profile: (str) use a separate account profile specified by name
        :save: (bool) save pertinent values in the messages config file,
            such as from_, acct_sid, auth_token (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :sid: (str) return value from send, record of sent message

    Managed Attributes (Properties):
        :auth: auth will set as a private attribute (_auth) and obscured when requested
        :from_: user input will be validated for a proper phone number
        :to: user input will be validated for a proper phone number
        :attachments: user input will be validated for a proper url

    Usage:
        Create a text message (SMS/MMS) object with required Args above.
        Send text message with self.send() or self.send_async() methods.

    Notes:
        For API description:
        https://www.twilio.com/docs/api/messaging/send-messages
    """

    auth = credential_property("auth")
    from_ = validate_property("from_")
    to = validate_property("to")
    attachments = validate_property("attachments")

    def __init__(
        self,
        from_=None,
        to=None,
        auth=None,
        body=" ",
        attachments=None,
        profile=None,
        save=False,
        verbose=False,
    ):

        self.from_ = from_
        self.to = to
        self.auth = auth
        self.body = body
        self.attachments = attachments
        self.profile = profile
        self.save = save
        self.verbose = verbose
        self.sid = None

        if self.profile:
            check_config_file(self)

    def __str__(self, indentation="\n"):
        """print(Email(**args)) method.
           Indentation value can be overridden in the function call.
           The default is new line"""
        return (
            "{}From: {}"
            "{}To: {}"
            "{}Body: {}"
            "{}Attachments: {}"
            "{}SID: {}".format(
                indentation,
                self.from_,
                indentation,
                self.to,
                indentation,
                reprlib.repr(self.body),
                indentation,
                self.attachments,
                indentation,
                self.sid,
            )
        )

    def send(self):
        """
        Send the SMS/MMS message.
        Set self.sid to return code of message.
        """
        url = (
            "https://api.twilio.com/2010-04-01/Accounts/"
            + self._auth[0]
            + "/Messages.json"
        )
        data = {
            "From": self.from_,
            "To": self.to,
            "Body": self.body,
            "MediaUrl": self.attachments,
        }

        if self.verbose:
            print(
                "Debugging info"
                "\n--------------"
                "\n{} Message created.".format(timestamp())
            )

        resp = requests.post(url, data=data, auth=(self._auth[0], self._auth[1]))
        self.sid = resp.json()["sid"]

        if self.verbose:
            print(
                timestamp(),
                type(self).__name__ + " info:",
                self.__str__(indentation="\n * "),
                "\n * HTTP status code:",
                resp.status_code,
            )

        if resp.status_code >= 200 and resp.status_code < 300:
            print("Message sent.")
        else:
            print("Error while sending.  HTTP status code =", resp.status_code)

    def send_async(self):
        """Send message asynchronously."""
        MESSAGELOOP.add_message(self)



"""
Module designed to make creating and sending chat messages easy.

1.  TelegramBot
    - Send messages via the Telegram Bot feature
    - https://core.telegram.org/bots
"""

import reprlib

import requests

from ._config import check_config_file
from ._eventloop import MESSAGELOOP
from ._interface import Message
from ._utils import credential_property
from ._utils import validate_property
from ._utils import timestamp


class TelegramBot(Message):
    """
    Create and send Telegram message via the Telegram Bot API.

    Args:
        :from_: (str) optional arg to specify who message is from.
        :auth: (str) auth token for bot for access.
        :chat_id: (str) chat_id for already-intiated chat.
            chat_id is an integer represented as a string.
            Recipient must have already initiated chat at some
            point in the past for bot to send message.
        :to: (str) if chat_id is unknown, can specify username of
            recipient to lookup via API call.  This may return None if
            chat is older than 24-hours old.
        :subject: (str) optional arg to specify message subject.
        :body: (str) message to send.
        :attachments: (str or list) each item is a url to attach
        :params: (dict) additional attributes to add to message,
            i.e. parse_mode (HTML or Markdown, see API for information
            on which attributes are possible.
        :profile: (str) use a separate account profile specified by name
        :save: (bool) save pertinent values in the messages config file,
            such as from_, chat_id, bot_token (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :message: (dict) current form of the message to be constructed

    Managed Attributes (Properties):
        :auth: auth will set as a private attribute (_auth) and obscured when requested
        :chat_id: user input will validate a proper integer as a string

    Usage:
        Create a TelegramBot object with required Args above.
        Send message with self.send() or self.send_async() methods.

    Note:
        For API description:
        https://core.telegram.org/bots/api#available-methods
    """

    auth = credential_property("auth")
    chat_id = validate_property("chat_id")

    def __init__(
        self,
        from_=None,
        auth=None,
        chat_id=None,
        to=None,
        subject=None,
        body="",
        attachments=None,
        params=None,
        profile=None,
        save=False,
        verbose=False,
    ):

        self.from_ = from_
        self.auth = auth
        self.chat_id = chat_id
        self.to = to
        self.subject = subject
        self.body = body
        self.attachments = attachments or []
        self.params = params or {}
        self.profile = profile
        self.save = save
        self.verbose = verbose
        self.message = {}

        if self.profile:
            check_config_file(self)

        self.base_url = "https://api.telegram.org/bot" + self._auth

    def __str__(self, indentation="\n"):
        """print(Telegram(**args)) method.
           Indentation value can be overridden in the function call.
           The default is new line"""
        return (
            "{}From: {}"
            "{}To: {}"
            "{}Chat ID: {}"
            "{}Subject: {}"
            "{}Body: {}"
            "{}Attachments: {}".format(
                indentation,
                self.from_,
                indentation,
                self.to,
                indentation,
                self.chat_id,
                indentation,
                self.subject,
                indentation,
                reprlib.repr(self.body),
                indentation,
                self.attachments,
            )
        )

    def get_chat_id(self, username):
        """Lookup chat_id of username if chat_id is unknown via API call."""
        if username is not None:
            chats = requests.get(self.base_url + "/getUpdates").json()
            user = username.split("@")[-1]
            for chat in chats["result"]:
                if chat["message"]["from"]["username"] == user:
                    return chat["message"]["from"]["id"]

    def _construct_message(self):
        """Build the message params."""
        self.message["chat_id"] = self.chat_id
        self.message["text"] = ""
        if self.from_:
            self.message["text"] += "From: " + self.from_ + "\n"
        if self.subject:
            self.message["text"] += "Subject: " + self.subject + "\n"

        self.message["text"] += self.body
        self.message.update(self.params)

    def _send_content(self, method="/sendMessage"):
        """send via HTTP Post."""
        if method == "/sendMessage":
            content_type = "Message body"
        elif method == "/sendDocument":
            content_type = "Attachment: " + self.message["document"]

        url = self.base_url + method
        resp = requests.post(url, json=self.message)

        if resp.status_code == 200 and self.verbose:
            print(timestamp(), content_type + " sent.")
        if resp.status_code > 300 and self.verbose:
            print(
                timestamp(), "Error while sending.  HTTP status code:", resp.status_code
            )
            print(resp.text)

    def send(self):
        """Start sending the message and attachments."""
        self._construct_message()
        if self.verbose:
            print(
                "Debugging info"
                "\n--------------"
                "\n{} Message created.".format(timestamp())
            )

        self._send_content("/sendMessage")

        for a in self.attachments:
            self.message["document"] = a
            self._send_content(method="/sendDocument")

        if self.verbose:
            print(
                timestamp(),
                type(self).__name__ + " info:",
                self.__str__(indentation="\n * "),
            )

        print("Message sent.")

    def send_async(self):
        """Send message asynchronously."""
        MESSAGELOOP.add_message(self)



"""
Module designed to make creating and sending chat messages easy.

1.  Slack
    - This is a base class that has the useful API for the other slack
      messages.
    - This is not to be instantiated on its own.

2.  SlackWebhook
    - Send messages via the Incoming Webhooks feature
    - https://api.slack.com/incoming-webhooks
    - Inherits functionality/API from the Slack class

3.  SlackPost
    - Send messages via the Slack chat.postMessage API
    - https://api.slack.com/methods/chat.postMessage
    - Inherits functionality/API from the Slack class
"""

import reprlib

import requests

from ._config import check_config_file
from ._eventloop import MESSAGELOOP
from ._interface import Message
from ._utils import credential_property
from ._utils import validate_property
from ._utils import timestamp


class Slack(Message):
    """Base class that Slack* classes inherit from."""

    def _construct_message(self):
        """Build the message params."""
        self.message["text"] = ""
        if self.from_:
            self.message["text"] += "From: " + self.from_ + "\n"
        if self.subject:
            self.message["text"] += "Subject: " + self.subject + "\n"

        self.message["text"] += self.body
        self._add_attachments()

    def _add_attachments(self):
        """Add attachments."""
        if self.attachments:
            if not isinstance(self.attachments, list):
                self.attachments = [self.attachments]

            self.message["attachments"] = [
                {"image_url": url, "author_name": ""} for url in self.attachments
            ]
            if self.params:
                for attachment in self.message["attachments"]:
                    attachment.update(self.params)

    def send(self, encoding="json"):
        """Send the message via HTTP POST, default is json-encoded."""
        self._construct_message()
        if self.verbose:
            print(
                "Debugging info"
                "\n--------------"
                "\n{} Message created.".format(timestamp())
            )

        if encoding == "json":
            resp = requests.post(self.url, json=self.message)
        elif encoding == "url":
            resp = requests.post(self.url, data=self.message)

        if self.verbose:
            print(
                timestamp(),
                type(self).__name__,
                " info:",
                self.__str__(indentation="\n * "),
                "\n * HTTP status code:",
                resp.status_code,
            )

        if resp.status_code >= 200 and resp.status_code < 300:
            print("Message sent.")
        else:
            print("Error while sending.  HTTP status code =", resp.status_code)

    def send_async(self):
        """Send message asynchronously."""
        MESSAGELOOP.add_message(self)


class SlackWebhook(Slack):
    """
    Create and send Slack message via the Incoming WebHooks API.

    Args:
        :from_: (str) optional arg to specify who message is from.
        :auth: (str) webhook url for installed slack app.
        :subject: (str) optional arg to specify message subject.
        :body: (str) message to send.
        :attachments: (str or list) each item is a url to attach
        :params: (dict) additional attributes to add to each attachment,
            i.e. author_name, title, text, etc., see API for information
            on which attributes are possible.
        :profile: (str) use a separate account profile specified by name
        :save: (bool) save pertinent values in the messages config file,
            such as from_, url (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :message: (dict) current form of the message to be constructed

    Managed Attributes (Properties):
        :auth: auth will set as a private attribute (_auth) and obscured when requested
        :attachments: user input will be validated for a proper url

    Usage:
        Create a SlackWebhook object with required Args above.
        Send message with self.send() or self.send_async() methods.

    Note:
        For API description:
        https://api.slack.com/incoming-webhooks
    """

    auth = credential_property("auth")
    attachments = validate_property("attachments")

    def __init__(
        self,
        from_=None,
        auth=None,
        subject=None,
        body="",
        attachments=None,
        params=None,
        profile=None,
        save=False,
        verbose=False,
    ):

        self.from_ = from_
        self.auth = auth
        self.subject = subject
        self.body = body
        self.attachments = attachments or []
        self.params = params
        self.profile = profile
        self.save = save
        self.verbose = verbose
        self.message = {}

        if self.profile:
            check_config_file(self)

        self.url = self._auth

    def __str__(self, indentation="\n"):
        """print(SlackWebhook(**args)) method.
           Indentation value can be overridden in the function call.
           The default is new line"""
        return (
            "{}URL: {}"
            "{}From: {}"
            "{}Subject: {}"
            "{}Body: {}"
            "{}Attachments: {}".format(
                indentation,
                self.url,
                indentation,
                self.from_ or "Not Specified",
                indentation,
                self.subject,
                indentation,
                reprlib.repr(self.body),
                indentation,
                self.attachments,
            )
        )


class SlackPost(Slack):
    """
    Create and send Slack message via the Slack chat.poseMessage API

    Args:
        :from_: (str) optional arg to specify who message is from.
        :auth: (str) authentication token with required scopes.
        :channel: (str) Channel, private group, or IM channel to send message
        :subject: (str) optional arg to specify message subject.
        :body: (str) message to send.
        :attachments: (str or list) each item is a url to attach
        :params: (dict) additional attributes to add to each attachment,
            i.e. author_name, title, text, etc., see API for information
            on which attributes are possible.
        :profile: (str) use a separate account profile specified by name
        :save: (bool) save pertinent values in the messages config file,
            such as from_, url (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :message: (dict) current form of the message to be constructed

    Managed Attributes (Properties):
        :auth: auth will set as a private attribute (_auth) and obscured when requested
        :attachments: user input will be validated for a proper url
        :channel: user input will be validated for a proper string

    Usage:
        Create a SlackPost object with required Args above.
        Send message with self.send() or self.send_async() methods.

    Note:
        For API description:
        https://api.slack.com/methods/chat.postMessage
    """

    auth = credential_property("auth")
    attachments = validate_property("attachments")
    channel = validate_property("channel")

    def __init__(
        self,
        from_=None,
        auth=None,
        channel=None,
        subject=None,
        body="",
        attachments=None,
        params=None,
        profile=None,
        save=False,
        verbose=False,
    ):

        self.from_ = from_
        self.auth = auth
        self.channel = channel
        self.subject = subject
        self.body = body
        self.attachments = attachments or []
        self.params = params
        self.profile = profile
        self.save = save
        self.verbose = verbose
        self.url = "https://slack.com/api/chat.postMessage"

        if self.profile:
            check_config_file(self)

        self.message = {"token": self._auth, "channel": self.channel}

    def __str__(self, indentation="\n"):
        """print(SlackPost(**args)) method.
           Indentation value can be overridden in the function call.
           The default is new line"""
        return (
            "{}Channel: {}"
            "{}From: {}"
            "{}Subject: {}"
            "{}Body: {}"
            "{}Attachments: {}".format(
                indentation,
                self.channel,
                indentation,
                self.from_ or "Not Specified",
                indentation,
                self.subject,
                indentation,
                reprlib.repr(self.body),
                indentation,
                self.attachments,
            )
        )

    def send(self):
        """Send the message via HTTP POST, url-encoded."""
        super().send(encoding="url")



"""
Module designed to make creating and sending emails easy.

1.  Email
    - Uses the Python 3 standard library MIMEMultipart email
      object to construct the email.
"""

import reprlib
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication

from ._config import check_config_file
from ._eventloop import MESSAGELOOP
from ._interface import Message
from ._utils import credential_property
from ._utils import validate_property
from ._utils import timestamp


SMTP_SERVERS = {
    "gmail.com": ("smtp.gmail.com", 465),
    "yahoo.com": ("smtp.yahoo.com", 465),
    "yahoo.co.uk": ("smtp.mail.yahoo.co.uk", 465),
    "yahoo.com.au": ("smtp.mail.yahoo.com.au", 465),
    "live.com": ("smtp.live.com", 465),
    "office365.com": ("smtp.office365.com", 587),
    "att.net": ("smtp.att.yahoo.com", 465),
    "verizon.net": ("outgoing.verizon.net", 465),
    "ntlworld.com": ("smtp.ntlworld.com", 465),
    "hotmail.com": ("smtp.live.com", 465),
    "mail.com": ("smtp.mail.com", 465),
}


class Email(Message):
    """
    Create and send emails using the built-in email package.

    Args:
        :from_: (str) originating email address
            i.e. 'me@here.com'
        :to: (str or list) primary message recipients
             i.e. 'you@there.com' or
                  ['her@there.com', 'him@there.com']
        :server: (str) url of smtp server
            i.e. 'smtp.gmail.com'
        :port: (int) smtp server port
            i.e. 465 or 587
        :auth: (str) password for email account
        :cc: (str or list) carbon-copy recipients
        :bcc: (str or list) blind carbon-copy recipients
        :subject: (str) email message subject line
        :body: (str) body text of the message to send
        :attachments: (str or list) files to attach
            i.e. './file1', or
                ['/home/you/file1.txt', '/home/you/file2.pdf']
        :profile: (str) use an account profile specified by name
        :save: (bool) save pertinent values in the messages config.json file,
            such as from_, server, port, auth (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :message: (MIMEMultipart) current form of the message to be constructed

    Managed Attributes (Properties):
        :auth: auth will set as a private attribute (_auth) and obscured when requested
        :from_: user input will validate a proper email address
        :to: user input will be validated for a proper email address
        :cc: user input will be validated for a proper email address
        :bcc: user input will be validated for a proper email address

    Usage:
        Create an email object with required Args above.
        Send email with self.send() or self.send_async() methods.

    Note:
        Some email servers may require you to modify security setting, such as
        gmail allowing "less secure apps" to access the account.  Otherwise
        failure may occur when attempting to send.
    """

    auth = credential_property("auth")
    from_ = validate_property("from_")
    to = validate_property("to")
    cc = validate_property("cc")
    bcc = validate_property("bcc")

    def __init__(
        self,
        from_=None,
        to=None,
        server=None,
        port=None,
        auth=None,
        cc=None,
        bcc=None,
        subject="",
        body="",
        attachments=None,
        profile=None,
        save=False,
        verbose=False,
    ):

        self.from_, self.to, self.cc, self.bcc = from_, to, cc, bcc
        self.server = server or self.get_server(from_)[0]
        self.port = port or self.get_server(from_)[1]
        self.auth = auth
        self.subject = subject
        self.body = body
        self.attachments = attachments or []
        self.profile = profile
        self.save = save
        self.verbose = verbose
        self.message = None

        if self.profile:
            check_config_file(self)

    def __str__(self, indentation="\n"):
        """print(Email(**args)) method.
           Indentation value can be overridden in the function call.
           The default is new line"""
        return (
            "{}Server: {}:{}"
            "{}From: {}"
            "{}To: {}"
            "{}Cc: {}"
            "{}Bcc: {}"
            "{}Subject: {}"
            "{}Body: {}"
            "{}Attachments: {}".format(
                indentation,
                self.server,
                self.port,
                indentation,
                self.from_,
                indentation,
                self.to,
                indentation,
                self.cc,
                indentation,
                self.bcc,
                indentation,
                self.subject,
                indentation,
                reprlib.repr(self.body),
                indentation,
                self.attachments,
            )
        )

    @staticmethod
    def get_server(address=None):
        """Return an SMTP servername guess from outgoing email address."""
        if address:
            domain = address.split("@")[1]
            try:
                return SMTP_SERVERS[domain]
            except KeyError:
                return ("smtp." + domain, 465)
        return (None, None)

    @staticmethod
    def list_to_string(recipient):
        """
        Format the recipient for the MIMEMultipart() email type.
        If the recipient is a list, then it returns the list as a
        comma separated string.
        example: input=['you@here.com', 'her@there.com']
                 output='you@there.com, her@there.com'
        """
        if recipient:
            if isinstance(recipient, list):
                return ", ".join(recipient)
            return recipient

    def _generate_email(self):
        """Put the parts of the email together."""
        self.message = MIMEMultipart()
        self._add_header()
        self._add_body()
        self._add_attachments()

    def _add_header(self):
        """Add email header info."""
        self.message["From"] = self.from_
        self.message["Subject"] = self.subject

    def _add_body(self):
        """Add body content of email."""
        if self.body:
            b = MIMEText("text", "plain")
            b.set_payload(self.body)
            self.message.attach(b)

    def _add_attachments(self):
        """Add required attachments."""
        num_attached = 0
        if self.attachments:
            if isinstance(self.attachments, str):
                self.attachments = [self.attachments]

            for item in self.attachments:
                doc = MIMEApplication(open(item, "rb").read())
                doc.add_header("Content-Disposition", "attachment", filename=item)
                self.message.attach(doc)
                num_attached += 1
        return num_attached

    def _get_session(self):
        """Start session with email server."""
        if self.port in (465, "465"):
            session = self._get_ssl()
        elif self.port in (587, "587"):
            session = self._get_tls()
        session.login(self.from_, self._auth)
        return session

    def _get_ssl(self):
        """Get an SMTP session with SSL."""
        return smtplib.SMTP_SSL(self.server, self.port)

    def _get_tls(self):
        """Get an SMTP session with TLS."""
        session = smtplib.SMTP(self.server, self.port)
        session.ehlo()
        session.starttls()
        session.ehlo()
        return session

    def send(self):
        """Send the message."""
        self._generate_email()

        if self.verbose:
            print(
                "Debugging info"
                "\n--------------"
                "\n{} Message created.".format(timestamp())
            )

        session = self._get_session()
        if self.verbose:
            print(timestamp(), "Login successful.")

        recipients = []
        if self.to:
            self.message["To"] = self.list_to_string(self.to)
            recipients.append(self.to)
        if self.cc:
            self.message["Cc"] = self.list_to_string(self.cc)
            recipients.append(self.cc)
        if self.bcc:
            self.message["Bcc"] = self.list_to_string(self.bcc)
            recipients.append(self.bcc)

        session.sendmail(self.from_, recipients, self.message.as_string())
        session.quit()

        if self.verbose:
            print(timestamp(), "Logged out.")

        if self.verbose:
            print(
                timestamp(),
                type(self).__name__ + " info:",
                self.__str__(indentation="\n * "),
            )

        print("Message sent.")

    def send_async(self):
        """Send message asynchronously."""
        MESSAGELOOP.add_message(self)



"""messages.cli tests."""

import builtins
import os
import sys

import pytest

import click
from click import Context
from click.testing import CliRunner

import messages.cli
from messages.cli import get_body_from_file
from messages.cli import trim_args
from messages.cli import send_message
from messages.cli import main
from messages.email_ import Email
from messages._exceptions import UnsupportedMessageTypeError

from conftest import skip_if_on_travisCI


##############################################################################
# FIXTURES
##############################################################################

class ClickConfig(Context):
    """Used to pass ctx for click applications."""
    def __init__(self): pass
    def get_usage(self): pass


@pytest.fixture()
def get_ctx():
    """Returns a mocked click.Context() object."""
    return ClickConfig()


@pytest.fixture()
def main_mocks(mocker):
    """Returns mocks for all funcs inside main()."""
    body_mk = mocker.patch.object(messages.cli, 'get_body_from_file')
    config_mk = mocker.patch.object(messages.cli, 'create_config_entry')
    send_mk = mocker.patch.object(messages.cli, 'send')
    return (body_mk, config_mk, send_mk)


##############################################################################
# TESTS: cli.get_body_from_file
##############################################################################

@skip_if_on_travisCI
def test_get_body_from_file(tmpdir):
    """
    GIVEN a call to messages via the CLI
    WHEN a message is specified by filename
    THEN assert the file contents are read and set into the message body
        * this uses the tmpdir pytest built-in fixture to create temporary
          directory and file for the test
    """
    msg_file = tmpdir.join('msg.txt')
    msg_file.write('This is the message to send!')
    kwds = {'body': None, 'file': msg_file}
    body = get_body_from_file(kwds)
    assert kwds == {'body': 'This is the message to send!', 'file': None}


def test_get_body_from_file_travis(mocker):
    """
    GIVEN a call to messages via the CLI
    WHEN a message is specified by filename
    THEN assert the correct file operations are called and the kwds
        dict is updated
    """
    isfile_mock = mocker.patch.object(os.path, 'isfile')
    open_mock = mocker.patch.object(builtins, 'open')
    isfile_mock.return_value = True
    kwds = {'file': 'file.txt'}
    get_body_from_file(kwds)
    assert kwds['file'] is None
    assert 'body' in kwds.keys()


##############################################################################
# TESTS: cli.trim_args
##############################################################################

def test_trim_args_rejectedKV():
    """
    GIVEN a call to messages via the CLI
    WHEN trim_args is called on the CLI args
    THEN assert the correct args are removed
    """
    kwds = {'type': 'email', 'arg1': None, 'arg2': (), 'arg3': 'valid'}
    kwargs = trim_args(kwds)
    assert kwargs == {'arg3': 'valid'}


def test_trim_args_ListItems():
    """
    GIVEN a call to messages via the CLI
    WHEN trim_args is called on the CLI args
    THEN assert values with keys to/cc/bcc/attach are returned as a list
        instead of the tuple generated by the click package
    """
    kwds = {'to': ('me@here.com',), 'cc': ('her@there.com',),
            'bcc': ('him@there.com', 'her@there.com'), 'attachments': ('file1',)}
    kwargs = trim_args(kwds)
    assert kwargs == {'to': ['me@here.com'], 'cc': ['her@there.com'],
            'bcc': ['him@there.com', 'her@there.com'], 'attachments': ['file1']}


##############################################################################
# TESTS: cli.send_message
##############################################################################

def test_send_message(mocker):
    """
    GIVEN a call to messages via the CLI
    WHEN send_message is called with a valid message type with
        kwds['file'] = None
    THEN assert the correct sequence is called
    """
    send_mock = mocker.patch.object(messages.cli, 'send')
    trim_mock = mocker.patch.object(messages.cli, 'trim_args')
    file_mock = mocker.patch.object(messages.cli, 'get_body_from_file')
    send_message('email', kwds={'file': None})
    assert trim_mock.call_count == 1
    assert send_mock.call_count == 1
    assert file_mock.call_count == 0


def test_send_message_withFile(mocker):
    """
    GIVEN a call to messages via the CLI
    WHEN send_message is called with a valid message type with
        kwds['file'] != None
    THEN assert the correct sequence is called
    """
    send_mock = mocker.patch.object(messages.cli, 'send')
    trim_mock = mocker.patch.object(messages.cli, 'trim_args')
    file_mock = mocker.patch.object(messages.cli, 'get_body_from_file')
    send_message('email', kwds={'file': 'Some File'})
    assert trim_mock.call_count == 1
    assert send_mock.call_count == 1
    assert file_mock.call_count == 1


##############################################################################
# TESTS: cli.main
##############################################################################

def test_main_configure(mocker):
    """
    GIVEN a call to messages via the CLI
    WHEN subcommand = configure
    THEN assert the correct sequence is called
    """
    config_mock = mocker.patch.object(messages.cli, 'create_config_profile')
    runner = CliRunner()
    runner.invoke(main, ['configure', 'email'], catch_exceptions=False)
    assert config_mock.call_count == 1


@pytest.mark.parametrize('subcommand',
                         ['email',
                          'twilio',
                          'slackwebhook',
                          'slackpost',
                          'telegram',
                          'whatsapp'
                         ])
def test_main_message(subcommand, mocker):
    """
    GIVEN a call to a messages via the CLI
    WHEN subcommand = one of the message types
    THEN assert the correct sequence is called
    """
    send_mock = mocker.patch.object(messages.cli, 'send_message')
    runner = CliRunner()
    runner.invoke(main, [subcommand, '--verbose'], catch_exceptions=False)
    assert send_mock.call_count == 1



"""Custom Exceptions Module."""


class InvalidMessageInputError(ValueError):
    """Exception for invalid inputs in message classes."""

    def __init__(self, msg_type, attr, value, input_type):
        self.err = "Invalid input for specified message class: " + msg_type
        self.err += "\n    * param: {}".format(attr)
        self.err += "\n    * value given: {!r}".format(value)
        self.err += "\n    * input type must be: {}".format(input_type)
        super(InvalidMessageInputError, self).__init__(self.err)


class UnsupportedMessageTypeError(TypeError):
    """Exception for declaring unsupported message types."""

    def __init__(self, msg_type, msg_types=None):
        self.err = "Invalid message type: " + msg_type
        if msg_types:
            self.err += "\n    * Supported message types: "
            self.err += str(msg_types)
        super(UnsupportedMessageTypeError, self).__init__(self.err)


class UnknownProfileError(KeyError):
    """Exception for unknown config.json profile names."""

    def __init__(self, profile):
        self.err = "Unknown Profile name: " + profile
        super(UnknownProfileError, self).__init__(self.err)



"""
Configuration module.

Handles the creation/editing of configuration files and saving of
default attributes, credentials (auth), etc.
"""

from collections import MutableSequence
from collections import OrderedDict
from getpass import getpass

import jsonconfig

from ._exceptions import UnknownProfileError


##############################################################################
#  Message class config defaults
##############################################################################

CONFIG = {
    "email": {
        "settings": {
            "from_": "From email address (e.g. you@here.com)",
            "server": "Email server url (e.g. smtp.here.com)",
            "port": "Email server port number (e.g. 465)",
        },
        "auth": {"auth": "Email service password"},
    },
    "slackwebhook": {
        "settings": {"from_": "Name or alias of sender (optional)"},
        "auth": {"auth": "Slack API Webhook URL"},
    },
    "slackpost": {
        "settings": {
            "from_": "Name or alias of sender (optional)",
            "channel": "Channel to post message to (e.g. #general)",
        },
        "auth": {"auth": "Slack API authentication token"},
    },
    "twilio": {
        "settings": {"from_": "Twilio phone number (e.g. +19998675309)"},
        "auth": OrderedDict(
            [
                ("auth_sid", "Twilio API account SID"),
                ("auth_token", "Twilio API authorization token"),
            ]
        ),
    },
    "telegrambot": {
        "settings": {"channel_id": "Telegram Channel ID of chat"},
        "auth": {"auth": "Telegram authorization token"},
    },
    "whatsapp": {
        "settings": {"from_": "Twilio phone number (e.g. +19998675309)"},
        "auth": OrderedDict(
            [
                ("auth_sid", "Twilio API account SID"),
                ("auth_token", "Twilio API authorization token"),
            ]
        ),
    },
}


##############################################################################
#  Config functions used by Message classes
##############################################################################


def check_config_file(msg):
    """
    Checks the config.json file for default settings and auth values.

    Args:
        :msg: (Message class) an instance of a message class.
    """
    with jsonconfig.Config("messages", indent=4) as cfg:
        verify_profile_name(msg, cfg)

        retrieve_data_from_config(msg, cfg)

        if msg._auth is None:
            retrieve_pwd_from_config(msg, cfg)

        if msg.save:
            update_config_data(msg, cfg)
            update_config_pwd(msg, cfg)


def verify_profile_name(msg, cfg):
    """
    Verifies the profile name exists in the config.json file.

    Args:
        :msg: (Message class) an instance of a message class.
        :cfg: (jsonconfig.Config) config instance.
    """
    if msg.profile not in cfg.data:
        raise UnknownProfileError(msg.profile)


def retrieve_data_from_config(msg, cfg):
    """
    Update msg attrs with values from the profile configuration if the
    msg.attr=None, else leave it alone.

    Args:
        :msg: (Message class) an instance of a message class.
        :cfg: (jsonconfig.Config) config instance.
    """
    msg_type = msg.__class__.__name__.lower()
    for attr in msg:
        if getattr(msg, attr) is None and attr in cfg.data[msg.profile][msg_type]:
            setattr(msg, attr, cfg.data[msg.profile][msg_type][attr])


def retrieve_pwd_from_config(msg, cfg):
    """
    Retrieve auth from profile configuration and set in msg.auth attr.

    Args:
        :msg: (Message class) an instance of a message class.
        :cfg: (jsonconfig.Config) config instance.
    """
    msg_type = msg.__class__.__name__.lower()
    key_fmt = msg.profile + "_" + msg_type
    pwd = cfg.pwd[key_fmt].split(" :: ")
    if len(pwd) == 1:
        msg.auth = pwd[0]
    else:
        msg.auth = tuple(pwd)


def update_config_data(msg, cfg):
    """
    Updates the profile's config entry with values set in each attr by the
    user.  This will overwrite existing values.

    Args:
        :msg: (Message class) an instance of a message class.
        :cfg: (jsonconfig.Config) config instance.
    """
    for attr in msg:
        if attr in cfg.data[msg.profile] and attr is not "auth":
            cfg.data[msg.profile][attr] = getattr(msg, attr)


def update_config_pwd(msg, cfg):
    """
    Updates the profile's auth entry with values set by the user.
    This will overwrite existing values.

    Args:
        :msg: (Message class) an instance of a message class.
        :cfg: (jsonconfig.Config) config instance.
    """
    msg_type = msg.__class__.__name__.lower()
    key_fmt = msg.profile + "_" + msg_type
    if isinstance(msg._auth, (MutableSequence, tuple)):
        cfg.pwd[key_fmt] = " :: ".join(msg._auth)
    else:
        cfg.pwd[key_fmt] = msg._auth


##############################################################################
#  Config functions used by the CLI to create profile entries
##############################################################################


def create_config_profile(msg_type):
    """
    Create a profile for the given message type.

    Args:
        :msg_type: (str) message type to create config entry.
    """
    msg_type = msg_type.lower()

    display_required_items(msg_type)

    if get_user_ack():
        profile_name = input("Profile Name: ")
        data = get_data_from_user(msg_type)
        auth = get_auth_from_user(msg_type)
        configure_profile(msg_type, profile_name, data, auth)


def display_required_items(msg_type):
    """
    Display the required items needed to configure a profile for the given
    message type.

    Args:
        :msg_type: (str) message type to create config entry.
    """
    print("Configure a profile for: " + msg_type)
    print("You will need the following information:")
    for k, v in CONFIG[msg_type]["settings"].items():
        print("   * " + v)
    print("Authorization/credentials required:")
    for k, v in CONFIG[msg_type]["auth"].items():
        print("   * " + v)


def get_user_ack():
    """
    Get the user's acknowledgement to continue.

    Returns: bool
    """
    ack = input("\nContinue [Y/N]? ")
    return ack in ("y", "Y")


def get_data_from_user(msg_type):
    """Get the required 'settings' from the user and return as a dict."""
    data = {}
    for k, v in CONFIG[msg_type]["settings"].items():
        data[k] = input(v + ": ")
    return data


def get_auth_from_user(msg_type):
    """Get the required 'auth' from the user and return as a dict."""
    auth = []
    for k, v in CONFIG[msg_type]["auth"].items():
        auth.append((k, getpass(v + ": ")))
    return OrderedDict(auth)


def configure_profile(msg_type, profile_name, data, auth):
    """
    Create the profile entry.

    Args:
        :msg_type: (str) message type to create config entry.
        :profile_name: (str) name of the profile entry
        :data: (dict) dict values for the 'settings'
        :auth: (dict) auth parameters
    """
    with jsonconfig.Config("messages", indent=4) as cfg:
        write_data(msg_type, profile_name, data, cfg)
        write_auth(msg_type, profile_name, auth, cfg)

    print("[+] Configuration entry for <" + profile_name + "> created.")
    print("[+] Configuration file location: " + cfg.filename)


def write_data(msg_type, profile_name, data, cfg):
    """
    Write the settings into the data portion of the cfg.

    Args:
        :msg_type: (str) message type to create config entry.
        :profile_name: (str) name of the profile entry
        :data: (dict) dict values for the 'settings'
        :cfg: (jsonconfig.Config) config instance.
    """
    if profile_name not in cfg.data:
        cfg.data[profile_name] = {}
    cfg.data[profile_name][msg_type] = data


def write_auth(msg_type, profile_name, auth, cfg):
    """
    Write the settings into the auth portion of the cfg.

    Args:
        :msg_type: (str) message type to create config entry.
        :profile_name: (str) name of the profile entry
        :auth: (dict) auth parameters
        :cfg: (jsonconfig.Config) config instance.
    """
    key_fmt = profile_name + "_" + msg_type
    pwd = []
    for k, v in CONFIG[msg_type]["auth"].items():
        pwd.append(auth[k])

    if len(pwd) > 1:
        cfg.pwd[key_fmt] = " :: ".join(pwd)
    else:
        cfg.pwd[key_fmt] = pwd[0]



"""messages.whatsapp tests."""

import pytest
import requests

import messages.whatsapp
from messages.whatsapp import WhatsApp
from messages.whatsapp import check_config_file
from messages._eventloop import MESSAGELOOP


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_whap(mocker):
    """Return a valid WhatsApp instance."""
    configure_mock = mocker.patch.object(messages.whatsapp, 'check_config_file')
    return WhatsApp(from_='+16198675309', to='+16195551212',
            auth=('test_sid', 'test_token'), body='test text!',
            attachments='https://imgs.xkcd.com/comics/python.png',
            profile='tester', save=False)


##############################################################################
# TESTS: WhatsApp.__init__
##############################################################################

def test_whatsapp_init(get_whap, cfg_mock):
    """
    GIVEN a need to create an WhatsApp object
    WHEN the user instantiates a new object with required args
    THEN assert WhatsApp object is created with given args
    """
    w = get_whap
    assert w.from_ == 'whatsapp:+16198675309'
    assert w.to == 'whatsapp:+16195551212'
    assert w.auth == '***obfuscated***'
    assert '_auth' in w.__dict__
    assert w._auth == ('test_sid', 'test_token')
    assert w.body == 'test text!'
    assert w.attachments == 'https://imgs.xkcd.com/comics/python.png'



import ast
import re
from setuptools import setup, find_packages


# get __version__ from __init__.py
_version_re = re.compile(r'__version__\s+=\s+(.*)')
with open('messages/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))


# load README.md
with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()


setup(
    name='messages',
    version=version,
    url='https://github.com/trp07/messages',
    keywords=['message', 'messages', 'wrapper', 'email', 'text', 'SMS', 'MMS',
              'chat', 'chats', 'slack', 'twilio', 'async', 'asynchronous',
              'telegram', 'bot', 'telegrambot'],

    author='Tim Phillips',
    author_email='phillipstr@gmail.com',

    description=('A package designed to make sending messages '
                 'easy and efficient!'),
    long_description=readme,
    long_description_content_type='text/markdown',

    packages=find_packages(include=['messages']),
    include_package_data=True,
    zipsafe=False,

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
    ],

    python_requires='>=3.5',

    install_requires=[
        'click>=6.0',
        'requests',
        'jsonconfig-tool',
        'validus>=0.3.0',
    ],

    test_suite='tests',
    test_requires=[
        'pytest-cov',
        'pytest-mock',
    ],

    setup_requires=['pytest-runner'],

    entry_points={
        'console_scripts': ['messages=messages.cli:main']
    },

)



"""This module implements the Messages API."""

from .email_ import Email
from .slack import SlackWebhook
from .slack import SlackPost
from .telegram import TelegramBot
from .text import Twilio
from .whatsapp import WhatsApp

from ._exceptions import UnsupportedMessageTypeError


MESSAGE_CLASSES = {Email, SlackWebhook, SlackPost, TelegramBot, Twilio, WhatsApp}

MESSAGE_TYPES = {i.__name__.lower(): i for i in MESSAGE_CLASSES}


def send(msg_type, send_async=False, *args, **kwargs):
    """
    Constructs a message class and sends the message.
    Defaults to sending synchronously.  Set send_async=True to send
    asynchronously.

    Args:
        :msg_type: (str) the type of message to send, i.e. 'Email'
        :send_async: (bool) default is False, set True to send asynchronously.
        :kwargs: (dict) keywords arguments that are required for the
            various message types.  See docstrings for each type.
            i.e. help(messages.Email), help(messages.Twilio), etc.

    Example:
        >>> kwargs = {
                  from_: 'me@here.com',
                  to: 'you@there.com',
                  password: 'yourPassword',
                  subject: 'Email Subject',
                  body: 'Your message to send',
                  attachments: ['filepath1', 'filepath2'],
            }
        >>> messages.send('email', **kwargs)
        Message sent...
    """
    message = message_factory(msg_type, *args, **kwargs)

    if send_async:
        message.send_async()
    else:
        message.send()


def message_factory(msg_type, msg_types=MESSAGE_TYPES, *args, **kwargs):
    """
    Factory function to return the specified message instance.

    Args:
        :msg_type: (str) the type of message to send, i.e. 'Email'
        :msg_types: (str, list, or set) the supported message types
        :kwargs: (dict) keywords arguments that are required for the
            various message types.  See docstrings for each type.
            i.e. help(messages.Email), help(messages.Twilio), etc.
    """
    try:
        return msg_types[msg_type.lower()](*args, **kwargs)
    except KeyError:
        raise UnsupportedMessageTypeError(msg_type, msg_types)
